import datetime
from typing import Callable, List, Optional

import humanize
import ipyvuetify as v
import mlflow
import solara
import traitlets


class SearchAndNewWidget(v.VuetifyTemplate):
    template_file = (__file__, "experiment_search.vue")

    search = traitlets.Unicode(allow_none=True).tag(sync=True)
    selected = traitlets.Any().tag(sync=True)
    items = traitlets.List().tag(sync=True)
    loading = traitlets.Bool(default_value=False).tag(sync=True)
    opened = traitlets.Bool(default_value=False).tag(sync=True)
    error = traitlets.Unicode(allow_none=True).tag(sync=True)


@solara.component
def SearchAndNew(
    search: Optional[str],
    selected: Optional[str],
    items: List[str],
    loading: bool,
    opened: bool,
    error: Optional[str],
    on_search: Callable[[str], None],
    on_selected: Callable[[str], None],
    on_opened: Callable[[bool], None],
):
    return SearchAndNewWidget.element(
        search=search,
        selected=selected,
        items=items,
        loading=loading,
        opened=opened,
        error=error,
        on_search=on_search,
        on_selected=on_selected,
        on_opened=on_opened,
    )


def date(time_in_millis):
    natural_time = humanize.naturaltime(datetime.datetime.now() - datetime.datetime.fromtimestamp(time_in_millis / 1000))
    return f"last updated {natural_time}"


@solara.component
def ExperimentSearch(value: Optional[str], on_value: Callable[[str], None]):
    search, set_search = solara.use_state("")
    opened, set_opened = solara.use_state(False)

    def get_experiments(_cancel):
        return mlflow.search_experiments()

    experiment_list_result: solara.Result = solara.use_thread(get_experiments, [])

    def get_all_items():
        items = (
            [
                {
                    "name": item.name,
                    "time": item.last_update_time,
                    "description": date(item.last_update_time),
                }
                for item in experiment_list_result.value
            ]
            if experiment_list_result.value
            else []
        )

        return sorted(items, key=lambda item: item["time"], reverse=True)

    all_items = solara.use_memo(get_all_items, [experiment_list_result.value])

    loading = experiment_list_result.state == solara.ResultState.RUNNING

    def filter_items():
        if search is None or search == "":
            return all_items
        else:
            return [item for item in all_items if search.lower() in item["name"].lower()]

    filtered_items = solara.use_memo(filter_items, [all_items, search])

    def on_selected(selected: str):
        set_opened(False)
        set_search("")
        on_value(selected)

    return SearchAndNew(
        search=search,
        selected=value,
        items=filtered_items,
        loading=loading,
        opened=opened,
        error=experiment_list_result.error.__repr__() if experiment_list_result.state == solara.ResultState.ERROR else None,
        on_search=set_search,
        on_selected=on_selected,
        on_opened=set_opened,
    )



from typing import List

import reacton.ipyvuetify as v
import solara


@solara.component
def Thumbnail(width: int = 600, height: int = 300, scale: float = 0.3, children: List = []):
    with v.Sheet(
        class_="domino-thumbnail",
        style_=f"height: calc({height}px * {scale}); width: calc({width}px * {scale}); overflow:hidden; position: relative; user-select: none;",
    ) as main:
        v.Sheet(
            style_=f"height: {height}px; width: {width}px; transform: scale({scale}); transform-origin: top left;",
            children=children,
        )
        solara.Div(style_="position: absolute; top: 0; left: 0; bottom: 0; right: 0")

    return main



import textwrap
from dataclasses import asdict, dataclass, field
from typing import Dict, List, Optional, Tuple

import plotly

from domino_code_assist import util
from domino_code_assist.util import in_dev_mode


@dataclass(frozen=True)
class Action:
    df_var_out: str = "df"


def pandas_fn_and_params(filename):
    fn_name = filename.lower().split(".")[-1]
    params = ""

    if fn_name.startswith("sas"):
        fn_name = "sas"
        params = ', encoding="utf-8"'
    elif fn_name == "xlsx" or fn_name == "xls":
        fn_name = "excel"
    elif fn_name == "tsv":
        fn_name = "csv"
        params = ', sep="\\t"'

    return fn_name, params


@dataclass(frozen=True)
class ActionOpen(Action):
    filename: Optional[str] = None
    reactive: bool = False

    def render_code(self, df_var_in):
        assert self.filename is not None

        fn_name, params = pandas_fn_and_params(self.filename)
        pd_part = f'pd.read_{fn_name}("{self.filename}"{params})'
        if not self.reactive:
            return f"import pandas as pd\n\n{self.df_var_out} = {pd_part}"

        return textwrap.dedent(
            f"""\
            import pandas as pd
            from domino_code_assist import app

            def load_{self.df_var_out}():
                return {pd_part}

            {self.df_var_out}_reactive = app.ReactiveDf("{self.df_var_out}", load_{self.df_var_out}, "drop dataframe {self.df_var_out} here")
            {self.df_var_out} = {self.df_var_out}_reactive.get()
            """
        )


@dataclass(frozen=True)
class ActionDemo(Action):
    df_name: str = "palmerpenguins"
    reactive: bool = False

    def render_code(self, df_var_in):
        if self.df_name in dir(plotly.data):
            import_part = "import plotly"
            pd_part = f"plotly.data.{self.df_name}()"
        else:
            import_part = "import domino_code_assist as dca"
            pd_part = f"dca.data.{self.df_name}()"

        if not self.reactive:
            return f"{import_part}\n\n{self.df_var_out} = {pd_part}"

        return textwrap.dedent(
            f"""\
            {import_part}
            from domino_code_assist import app

            def load_{self.df_var_out}():
                return {pd_part}

            {self.df_var_out}_reactive = app.ReactiveDf("{self.df_var_out}", load_{self.df_var_out}, "drop dataframe {self.df_var_out} here")
            {self.df_var_out} = {self.df_var_out}_reactive.get()
            """
        )


@dataclass(frozen=True)
class ActionUseNbLocals(Action):
    var_name: Optional[str] = None

    def render_code(self, df_var_in):
        assert self.var_name is not None
        return None


@dataclass(frozen=True)
class ActionDownloadDataSource(Action):
    name: Optional[str] = None
    type_: Optional[str] = None
    use_query: bool = False
    query: Optional[str] = None
    database: Optional[str] = None
    schema: Optional[str] = None
    table: Optional[str] = None
    sample: bool = False
    project: Optional[str] = None
    region: Optional[str] = None

    def render_code(self, df_var_in):
        package = "domino_code_assist.domino_lab_mock" if in_dev_mode() else "domino.data_sources"

        if self.type_ in ["S3Config", "GCSConfig"]:
            assert self.database
            s3_key = self.database
            fn_name, params = pandas_fn_and_params(s3_key)
            return textwrap.dedent(
                f"""\
                import io
                import pandas as pd
                from {package} import DataSourceClient

                ds = DataSourceClient().get_datasource("{self.name}")

                data = io.BytesIO(ds.get("{s3_key}"))
                {self.df_var_out} = pd.read_{fn_name}(data{params})
                """
            )

        # using limit instead of sample for now, since it's much faster for very large tables. " sample row (10000 rows)"
        query = self.query if self.use_query else f"select * from {self.database}.{self.schema}.{self.table}" + (" limit 2000" if self.sample else "")
        if self.type_ == "BigQueryConfig" and not self.use_query:
            query = f"select * from `{self.project + '.' if self.project else ''}{self.schema}.{self.table}`" + (" limit 2000" if self.sample else "")

        if self.database and self.schema and self.use_query:
            if self.type_ == "SnowflakeConfig":
                return textwrap.dedent(
                    f'''\
                    from {package} import DataSourceClient, SnowflakeConfig

                    ds = DataSourceClient().get_datasource("{self.name}")
                    ds.update(SnowflakeConfig(database="{self.database}", schema="{self.schema}"))

                    res = ds.query("""{query} """)
                    {self.df_var_out} = res.to_pandas()
                    '''
                )

        return textwrap.dedent(
            f'''\
            from {package} import DataSourceClient

            ds = DataSourceClient().get_datasource("{self.name}")
            res = ds.query("""{query} """)
            {self.df_var_out} = res.to_pandas()
            '''
        )


def quote(value):
    if '"' in value:
        if "'" not in value:
            return f"'{value}'"
        if '"""' not in value and not value.endswith('"'):
            return f'"""{value}"""'
        if "'''" not in value and not value.endswith("'"):
            return f"'''{value}'''"
        else:
            return repr(value)
    else:
        return f'"{value}"'


@dataclass(frozen=True)
class ActionFilter(Action):
    col: Optional[str] = None
    dtype: Optional[str] = None
    op: Optional[str] = None
    value: Optional[str] = None
    is_string: bool = False

    def render_code(self, df_var_in):
        if not self.is_string and self.value is not None and self.value.lower() == "nan":
            fn = "notna" if self.op == "!=" else "isna"
            return f'{self.df_var_out} = {df_var_in}.loc[{df_var_in}["{self.col}"].{fn}()]'

        value = self.value
        if self.is_string:
            value = quote(value)

        return f'{self.df_var_out} = {df_var_in}.loc[{df_var_in}["{self.col}"] {self.op} {value}]'

    def is_valid(self):
        return self.col is not None and self.op is not None and self.value is not None and self.df_var_out


@dataclass(frozen=True)
class ActionSelectColumns(Action):
    columns: Optional[List[str]] = None

    def render_code(self, df_var_in):
        assert self.columns
        col_str = ", ".join([f'"{c}"' for c in self.columns])
        return f"{self.df_var_out} = {df_var_in}[[{col_str}]]"

    def is_valid(self):
        return bool(self.columns and self.df_var_out)


@dataclass(frozen=True)
class ActionDropColumns(Action):
    columns: Optional[List[str]] = None

    def render_code(self, df_var_in):
        assert self.columns
        col_str = ", ".join([f'"{c}"' for c in self.columns])
        return f"{self.df_var_out} = {df_var_in}.drop(columns=[{col_str}])"

    def is_valid(self):
        return bool(self.columns and self.df_var_out)


@dataclass(frozen=True)
class ActionGroupBy(Action):
    columns: Optional[List[str]] = None
    aggregations: Optional[List[Tuple[str, str]]] = None
    df_var_out: str = "dfg"

    def render_code(self, df_var_in):
        if not self.columns or not self.aggregations:
            raise Exception("Action not complete")
        col_str = ", ".join([f'"{c}"' for c in self.columns])
        agg_str = ", ".join([f'{util.python_safe_name(name)}_{agg}=("{name}", "{agg}")' for name, agg in self.aggregations])
        code = f"{df_var_in}.groupby([{col_str}]).agg({agg_str}).reset_index()"
        return f"{self.df_var_out} = {code}"

    def is_valid(self):
        return bool(self.columns and self.df_var_out)


@dataclass(frozen=True)
class ActionMerge(Action):
    on_index: bool = False
    suffix: str = "_right"
    auto: bool = True
    var_name_other: Optional[str] = None
    mapping: Dict[Optional[str], Optional[str]] = field(default_factory=lambda: {None: None})
    how: Optional[str] = None

    def render_code(self, df_var_in):
        merge_arg = f", how='{self.how}'" if self.how else ""
        if self.on_index:
            return f"{self.df_var_out} = {df_var_in}.join({self.var_name_other}, rsuffix='{self.suffix}')"
        if self.auto:
            return f"{self.df_var_out} = {df_var_in}.merge({self.var_name_other}{merge_arg})"
        else:
            mapping = self.complete_mapping
            args = f"on={list(mapping.keys())}"
            has_renames = bool({k: v for k, v in mapping.items() if k != v})
            if has_renames:
                args = f"left_on={list(mapping.keys())}, right_on={list(mapping.values())}"
            return f"{self.df_var_out} = {df_var_in}.merge({self.var_name_other}, {args}, suffixes=(None, '{self.suffix}'){merge_arg})"

    def is_valid(self):
        return bool(self.var_name_other and self.df_var_out and (self.on_index or self.auto or self.complete_mapping))

    @property
    def complete_mapping(self):
        return {k: v for k, v in self.mapping.items() if k is not None and v is not None}


def render_code_chunks(actions):
    last_out = None
    chunks = []
    for action in actions:
        code_line = action.render_code(last_out)
        if code_line:
            chunks.append(code_line)
        last_out = action.df_var_out

    chunks.append(action.df_var_out)
    return chunks


def render_code(chunks):
    return "\n".join(chunks)


def to_json(obj):
    if isinstance(obj, Action):
        return {"type_action": {"name": obj.__class__.__name__, "data": asdict(obj)}}


def as_action(dct):
    if "type_action" in dct:
        dct = dct["type_action"]
        return globals()[dct["name"]](**dct["data"])
    return dct


def from_json(dct):
    if "type_action" in dct:
        dct = dct["type_action"]
        return globals()[dct["name"]](**dct["data"])



def palmerpenguins():
    import os

    import pandas

    return pandas.read_csv(os.path.join(os.path.dirname(__file__), "penguins.csv")).drop(columns=["rowid"])


def small_molecule_drugbank():
    import os

    import pandas

    return pandas.read_csv(os.path.join(os.path.dirname(__file__), "small_molecule_drugbank.csv")).drop(columns=["Unnamed: 0"])


def stocks_long():
    import os

    import pandas

    return pandas.read_parquet(os.path.join(os.path.dirname(__file__), "stocks_long.parquet"))



from typing import Callable, List, Optional

import ipyvuetify as vy
import ipywidgets
import reacton
import solara as sol
import traitlets

from domino_code_assist import util


class DrawerWidget(vy.VuetifyTemplate):
    template_file = (__file__, "drawer.vue")

    is_open = traitlets.Bool(False).tag(sync=True)
    content = traitlets.Any().tag(sync=True, **ipywidgets.widget_serialization)
    title = traitlets.Unicode().tag(sync=True)
    on_apply = traitlets.Callable(default_value=lambda: None, allow_none=True)
    show_default_buttons = traitlets.Bool(True).tag(sync=True)
    edit = traitlets.Bool(False).tag(sync=True)
    apply_disabled = traitlets.Bool(False).tag(sync=True)
    show_var_out = traitlets.Bool(False).tag(sync=True)
    var_out = traitlets.Unicode().tag(sync=True)
    valid_variable_name = traitlets.Bool(True).tag(sync=True)
    width = traitlets.Unicode("unset").tag(sync=True)
    warning_widget = traitlets.Any().tag(sync=True, **ipywidgets.widget_serialization)
    class_ = traitlets.Unicode("").tag(sync=True)

    def vue_apply(self, *ignored):
        self.on_apply and self.on_apply()


@reacton.component
def RightDrawer(
    open: bool,
    on_open: Callable[[bool], None],
    title: str,
    children: List[reacton.core.Element] = None,
    show_default_buttons: bool = True,
    on_apply: Optional[Callable[[], None]] = None,
    edit: bool = False,
    apply_disabled: bool = False,
    show_var_out: bool = False,
    var_out: str = "",
    on_var_out: Optional[Callable[[str], None]] = lambda *_: None,
    width: str = "unset",
    warning_widget: Optional[reacton.core.Element] = None,
    class_: str = "",
):
    valid_variable_name = util.is_valid_variable_name(var_out)

    content = None
    if children and len(children) == 1:
        content = children[0]
    elif children:
        content = sol.Div(children=children)
    return DrawerWidget.element(
        is_open=open,
        on_is_open=on_open,
        content=content,
        title=title,
        show_default_buttons=show_default_buttons,
        on_apply=on_apply,
        edit=edit,
        apply_disabled=(show_var_out and not valid_variable_name) or apply_disabled,
        show_var_out=show_var_out,
        var_out=var_out,
        on_var_out=on_var_out,
        valid_variable_name=valid_variable_name,
        width=width,
        warning_widget=warning_widget,
        class_=class_,
    )



import ipyvuetify as v
import reacton
import traitlets
from pygments import highlight
from pygments.formatters import HtmlFormatter
from pygments.lexers import PythonLexer


class CodeWidget(v.VuetifyTemplate):
    template_file = (__file__, "code.vue")

    highlighted_chunks = traitlets.List().tag(sync=True)
    error = traitlets.List(allow_none=True).tag(sync=True)
    no_controls = traitlets.Bool().tag(sync=True)
    assistant_mode = traitlets.Bool().tag(sync=True)

    event = traitlets.Any().tag(sync=True)


@reacton.component
def Code(code_chunks, on_event, error=None, no_controls=False, assistant_mode=False):
    def to_code_chunks(plain_chunks):
        formatter = HtmlFormatter()
        lexer = PythonLexer()
        return [highlight(chunk, lexer, formatter) for chunk in plain_chunks]

    return CodeWidget.element(
        highlighted_chunks=to_code_chunks(code_chunks), on_event=on_event, error=error, no_controls=no_controls, assistant_mode=assistant_mode
    )



from typing import TYPE_CHECKING, List, Optional

import reacton
import reacton.ipyvuetify as v
import solara

from .state import ReactiveDataSource, load_data

if TYPE_CHECKING:
    # needed for python 3.6 support
    from domino_data.data_sources import ObjectStoreDatasource


supported_types = ["S3Config", "GCSConfig"]


@reacton.component
def Panel(ds: "ObjectStoreDatasource"):
    data_source: ReactiveDataSource = load_data.data_source
    data_source.use()

    def get_s3_keys(_cancel) -> Optional[List]:
        if ds and ds.datasource_type in supported_types:
            res = ds.list_objects()
            return res
        else:
            return None

    s3_keys_result: solara.Result[Optional[List]] = solara.hooks.use_thread(get_s3_keys, [ds])

    with v.Sheet() as main:
        if s3_keys_result.error:
            solara.Warning(str(s3_keys_result.error))
        elif s3_keys_result.value is not None:
            keys = [obj.key for obj in s3_keys_result.value]
            with v.List():

                def on_value(index):
                    data_source.set_database(keys[index] if index is not None else None)

                database = data_source.value.database
                with v.ListItemGroup(v_model=keys.index(database) if database else None, on_v_model=on_value).meta(ref="s3"):
                    for key in keys:
                        with v.ListItem().key(key):
                            with v.ListItemIcon():
                                v.Icon(children=["mdi-file-document"])
                            with v.ListItemContent():
                                v.ListItemTitle(children=key)
        else:
            v.ProgressLinear(indeterminate=True).meta(ref="s3_keys_loading")

    return main



import threading
import time

import ipyvuetify as v
import solara
import traitlets
from ipywidgets import widget_serialization

from domino_code_assist.domino_api import get_domino_api
from domino_code_assist.settings import settings


class AboveOutputWidget(v.VuetifyTemplate):
    template_file = (__file__, "above_output.vue")
    children = traitlets.List().tag(sync=True, **widget_serialization)


@solara.component
def OpenExperiment(name: str):
    def get_experiments(cancel: threading.Event):
        for i in range(10):
            if cancel.is_set():
                raise Exception("Cancelled")
            found = [e for e in get_domino_api().get_experiments() if e["name"] == name]
            if cancel.is_set():
                raise Exception("Cancelled")
            if found:
                return found[0]
            time.sleep(1)
        raise Exception(f"Experiment {name} not found")

    experiments_result: solara.Result = solara.use_thread(get_experiments, [name])

    with AboveOutputWidget.element():
        if experiments_result.state == solara.ResultState.ERROR:
            solara.Error(f"Error getting link to experiment. Reason: {experiments_result.error.__repr__()}").key("error")
        else:
            finished = experiments_result.state == solara.ResultState.FINISHED

            if not experiments_result.value:
                solara.Error("Could not find experiment.")
            else:
                with solara.Div().key("found"):
                    href = (
                        f"/experiments/{settings.domino_project_owner}/{settings.domino_project_name}/{experiments_result.value['experiment_id']}"
                        if finished
                        else ""
                    )
                    solara.Button(
                        f"open experiment: {name}",
                        disabled=not finished,
                        color="primary",
                        icon_name="mdi-open-in-new",
                        href=href,
                        target="_blank",
                        style_="text-transform: unset",
                    )



import shutil
from pathlib import Path

import playwright
from playwright.sync_api import Page

from domino_code_assist.playwright import mouse_move_middle

from .assistant import AssistentHelper

timeout = 16 * 1000
HERE = Path(__file__)


class Notebook:
    page: Page

    def __init__(self, page: Page, host: str = "localhost", port: int = 11112):
        self.host = host
        self.port = port
        # # TODO: this should actually happen at the server
        self.page = page
        shutil.copy(
            HERE.parent.parent.parent / "tests/e2e/notebooks/empty.ipynb",
            HERE.parent.parent.parent / "tests/e2e/notebooks/working.ipynb",
        )
        url = f"http://{self.host}:{self.port}/notebooks/working.ipynb"
        self.page.goto(url)
        self.page.set_default_timeout(timeout=timeout)
        # at least wait till the page has loaded some frontend
        self.page.locator("text=Domino Code Assist").wait_for()

        # if we click the kernellink too fast, it can close again
        self.page.wait_for_timeout(1000)
        self.page.locator("text=Kernel starting, please wait...").wait_for(state="detached")
        self.page.wait_for_timeout(1000)

        self.page.locator("#kernellink").click()
        self.page.locator('span:has-text("Restart")').click()
        self.page.locator('button:has-text("Restart")').click()

        self.page.locator("Kernel Ready").wait_for(state="detached")

    @property
    def assistant(self):
        return AssistentHelper(self.page)

    def last_cell_assistant_hover(self):
        last_input = self.last_code_cell.locator(".input")
        last_input.scroll_into_view_if_needed()
        mouse_move_middle(self.page, last_input)
        self.assistant.domino_logo.wait_for()
        self.page.wait_for_timeout(100)
        self.assistant.domino_logo.hover()
        self.page.wait_for_timeout(100)

    @property
    def code_cell(self):
        return self.page.locator(".code_cell")

    @property
    def last_code_cell(self):
        self.page.locator("text=In [ ]:").wait_for()
        return self.page.locator(".code_cell").last

    def insert_code(self, code, delay=0):
        # if we don't want, we sometimes jump to 'command' mode
        self.page.wait_for_timeout(500)
        input = self.page.locator("textarea").last
        input.focus()
        self.page.wait_for_timeout(100)
        input.type(code, delay=delay)

        input = self.page.locator("text=In [ ]:").last
        input.press("Shift+Enter")
        try:
            self.page.locator("css=.kernel_busy_icon").wait_for(timeout=1000)
        except playwright._impl._api_types.TimeoutError:
            pass
        self.page.locator("css=.kernel_busy_icon").wait_for(state="detached")

    def load_csv(self, path: Path, df_var_name: str = "df"):
        self.insert_code(
            f"""
import pandas as pd
{df_var_name} = pd.read_csv("{str(path)}")
{df_var_name}.head(2)""".strip()
        )



from dataclasses import replace
from typing import Callable, List, Type, Union, cast

import pandas as pd
import reacton
import reacton.ipyvuetify as v
import solara as sol
from solara import Tooltip

from domino_code_assist import action
from domino_code_assist.util import is_valid_variable_name

from .. import util


@reacton.component
def FilterPanel(columns, dtypes, on_action, fill):
    state, set_state = reacton.use_state(fill)

    def make_action():
        on_action(state)

    with sol.Div().meta(ref="FilterPanel") as panel:
        with sol.Div():
            FilterPanelView(columns, dtypes, state, set_state)

        with v.CardActions():
            v.Spacer()
            sol.Button(
                "apply",
                color="primary",
                icon_name="mdi-check",
                disabled=not state.is_valid(),
                on_click=make_action,
            )
    return panel


def escape_special_chars(s):
    return s.replace("\n", "\\n").replace("\t", "\\t").replace("\r", "\\r") if s else s


def unescape_special_chars(s):
    return s.replace("\\n", "\n").replace("\\t", "\t").replace("\\r", "\r") if s else s


@reacton.component
def FilterPanelView(columns, dtypes, action_: action.ActionFilter, on_filter_action):
    with sol.Div().meta(ref="FilterPanel") as main:
        with v.Row():
            with v.Col():
                v.Select(label="Column", v_model=action_.col, items=columns, on_v_model=lambda v: on_filter_action(replace(action_, col=v, dtype=dtypes[v])))
            with v.Col():
                v.Select(
                    label="Operator",
                    v_model=action_.op,
                    items=["<", ">", "<=", ">=", "!=", "=="],
                    on_v_model=lambda v: on_filter_action(replace(action_, op=v)),
                )
            with v.Col():

                def is_float_or_nan(value):
                    if value.lower() == "nan":
                        return True
                    try:
                        float(value)
                        return True
                    except ValueError:
                        return False

                v.TextField(
                    label="Value",
                    v_model=escape_special_chars(action_.value),
                    on_v_model=lambda v: on_filter_action(replace(action_, value=unescape_special_chars(v))),
                    error_messages=['Invalid value, consider using "as string"']
                    if action_.value and not action_.is_string and not is_float_or_nan(action_.value)
                    else None,
                )
            with v.Col():
                v.Switch(label="as string", v_model=action_.is_string, on_v_model=lambda v: on_filter_action(replace(action_, is_string=v)))
        with v.Row():
            with v.Col(sm=4):
                valid_variable_name = is_valid_variable_name(action_.df_var_out)
                v.TextField(
                    label="New dataframe name",
                    v_model=action_.df_var_out,
                    on_v_model=lambda v: on_filter_action(replace(action_, df_var_out=v)),
                    error_messages=[] if valid_variable_name else ["Invalid variable name"],
                )

    return main


@reacton.component
def SelectColumnsPanelView(columns, dtypes, action_: Union[action.ActionSelectColumns, action.ActionDropColumns], on_columns_action):
    with sol.Div().meta(ref="SelectColumnsPanel") as main:
        with v.Row():
            with v.Col():
                v.Select(
                    label="Columns",
                    v_model=action_.columns,
                    items=columns,
                    on_v_model=lambda v: on_columns_action(replace(action_, columns=v)),
                    multiple=True,
                    deletable_chips=True,
                )
        with v.Row():
            with v.Col():
                valid_variable_name = is_valid_variable_name(action_.df_var_out)
                v.TextField(
                    label="New dataframe name",
                    v_model=action_.df_var_out,
                    on_v_model=lambda v: on_columns_action(replace(action_, df_var_out=v)),
                    error_messages=[] if valid_variable_name else ["Invalid variable name"],
                )
    return main


@reacton.component
def GroupByPanelView(columns, dtypes, action_: action.ActionGroupBy, on_group_by_action):
    with sol.Div().meta(ref="GroupByPanel") as main:
        with v.Row():
            with v.Col(sm=4):
                v.Select(
                    label="Columns to group by",
                    v_model=action_.columns,
                    items=columns,
                    on_v_model=lambda v: on_group_by_action(replace(action_, columns=v)),
                    multiple=True,
                    deletable_chips=True,
                )
            with v.Col(sm=8, class_="pa-0"):
                new_aggs = []
                i = 0
                agg_names = ["size", "sum", "mean", "min", "max"]
                # filter out empty entries
                aggs = action_.aggregations and [k for k in action_.aggregations if not all(el is None for el in k)]
                for col, agg in aggs or []:
                    with v.Row():
                        with v.Col():
                            colnew = sol.ui_dropdown("Column to aggregate", col, columns, key=f"col_{i}")
                        with v.Col():
                            aggnew = sol.ui_dropdown("Aggregator", agg, agg_names, key=f"agg_{i}")
                            new_aggs.append((colnew, aggnew))
                            i += -1

                with v.Row():
                    with v.Col():
                        col = sol.ui_dropdown("Column to aggregate", None, columns, key=f"col_{i}")
                    with v.Col():
                        agg_name = sol.ui_dropdown("Aggregator", None, agg_names, key=f"agg_{i}")
                        if col and agg_name:
                            new_aggs.append((col, agg_name))

                on_group_by_action(replace(action_, aggregations=new_aggs))
        with v.Row():
            with v.Col(sm=4):
                valid_variable_name = is_valid_variable_name(action_.df_var_out)
                v.TextField(
                    label="New dataframe name",
                    v_model=action_.df_var_out,
                    on_v_model=lambda v: on_group_by_action(replace(action_, df_var_out=v)),
                    error_messages=[] if valid_variable_name else ["Invalid variable name"],
                )

    return main


@reacton.component
def MergePanelView(columns, dtypes, action_: action.ActionMerge, on_merge_action):
    dfs = util.get_vars(lambda v: isinstance(v, pd.DataFrame))

    common_cols, set_common_cols = sol.use_state(cast(List[str], []))

    def get_common_cols():
        if not action_.var_name_other:
            return
        cols = util.nb_locals[action_.var_name_other].columns.tolist()
        common = list(set(columns).intersection(set(cols)))

        mapping = {k: v for k, v in action_.mapping.items() if k is not None and v is not None}
        missing_mappings = [k for k in mapping.keys() if k not in columns] + [v for v in mapping.values() if v not in cols]
        if not mapping or missing_mappings:
            on_merge_action(replace(action_, auto=bool(common), mapping={None: None}))
        set_common_cols(common)

    sol.use_effect(get_common_cols, [columns, action_.var_name_other])

    def handle_empty(changed):
        incomplete = [k for k, v in changed.items() if k is None or v is None]
        if not incomplete:
            return {**changed, None: None}
        return changed

    def to_item(df_name):
        cols = util.nb_locals[df_name].columns.tolist()
        matching_cols = ", ".join(set(columns).intersection(set(cols)))
        return {"text": f"{df_name} [{matching_cols}]", "value": df_name}

    items = [to_item(df) for df in dfs]

    def update_key(new, old):
        changed = {new if k == old else k: v for k, v in action_.mapping.items()}
        on_merge_action(replace(action_, mapping=handle_empty(changed)))

    def update_value(new, key):
        changed = action_.mapping.copy()
        changed[key] = new
        on_merge_action(replace(action_, mapping=handle_empty(changed)))

    def remove(key):
        changed = action_.mapping.copy()
        del changed[key]
        on_merge_action(replace(action_, mapping=handle_empty(changed)))

    how_options = [None, "left", "right", "outer"]

    with v.Sheet():
        v.Select(
            label="Dataframe to join/merge",
            items=items,
            v_model=action_.var_name_other,
            on_v_model=lambda v: on_merge_action(replace(action_, var_name_other=v)),
        )

        if action_.var_name_other:
            v.Switch(v_model=action_.on_index, on_v_model=lambda v: on_merge_action(replace(action_, on_index=v)), label="Join on index")
            if not action_.on_index:
                with v.BtnToggle(
                    v_model=how_options.index(action_.how), on_v_model=lambda v: on_merge_action(replace(action_, how=how_options[v])), mandatory=True
                ):
                    with Tooltip("Inner (default)"):
                        sol.Button(icon_name="mdi-set-center")
                    with Tooltip("Left outer"):
                        sol.Button(icon_name="mdi-set-left-center")
                    with Tooltip("Right outer"):
                        sol.Button(icon_name="mdi-set-center-right")
                    with Tooltip("Full outer"):
                        sol.Button(icon_name="mdi-set-all")

                v.Switch(
                    v_model=action_.auto,
                    on_v_model=lambda v: on_merge_action(replace(action_, auto=v)),
                    label="Join on all common columns",
                    disabled=not common_cols,
                )
                if action_.auto:
                    v.Html(tag="span", children=["Join on column(s): "])
                    v.Html(tag="span", children=[", ".join(common_cols)])
                else:
                    left_cols = [c for c in columns if c not in action_.mapping]
                    right_cols = [c for c in util.nb_locals[action_.var_name_other].columns.tolist() if c not in action_.mapping.values()]
                    for map_a, map_b in action_.mapping.items():
                        with v.Row(style_="align-items: center"):
                            with v.Col():
                                sol.Div(children=[" join "])
                            with v.Col():
                                v.Select(
                                    label="column",
                                    items=([map_a] if map_a else []) + left_cols,
                                    v_model=map_a,
                                    on_v_model=lambda new, old=map_a: update_key(new, old),
                                )
                            with v.Col():
                                sol.Div(children=[" on "])
                            with v.Col():
                                v.Select(
                                    label="column",
                                    items=([map_b] if map_b else []) + right_cols,
                                    v_model=map_b,
                                    on_v_model=lambda new, key=map_a: update_value(new, key),
                                )
                            with v.Col():
                                if not (map_a is None and map_b is None):
                                    sol.Button(icon_name="mdi-trash-can", icon=True, on_click=lambda key=map_a: remove(key))
        valid_variable_name = is_valid_variable_name(action_.df_var_out)
        v.TextField(
            label="New dataframe name",
            v_model=action_.df_var_out,
            on_v_model=lambda v: on_merge_action(replace(action_, df_var_out=v)),
            error_messages=[] if valid_variable_name else ["Invalid variable name"],
        )


@reacton.component
def TransformActionSelectPanel(
    columns,
    dtypes,
    action_: action.Action,
    on_action,
    on_action_type: Callable[
        [Type[Union[action.ActionFilter, action.ActionSelectColumns, action.ActionDropColumns, action.ActionGroupBy, action.ActionMerge]]], None
    ],
):
    action_mapping = {
        action.ActionFilter: "Filter rows",
        action.ActionDropColumns: "Drop columns",
        action.ActionSelectColumns: "Select columns",
        action.ActionGroupBy: "Groupby and aggregate",
        action.ActionMerge: "Join/merge",
    }
    action_mapping_reverse = {val: k for k, val in action_mapping.items()}

    op = action_mapping[type(action_)]

    def get_panel():
        if isinstance(action_, action.ActionFilter):
            return FilterPanelView(columns, dtypes, action_, on_action).key("filter_rows")
        if isinstance(action_, action.ActionDropColumns):
            return SelectColumnsPanelView(columns, dtypes, action_, on_action).key("drop columns")
        if isinstance(action_, action.ActionSelectColumns):
            return SelectColumnsPanelView(columns, dtypes, action_, on_action).key("select columns")
        if isinstance(action_, action.ActionGroupBy):
            return GroupByPanelView(columns, dtypes, action_, on_action).key("group by")
        if isinstance(action_, action.ActionMerge):
            return MergePanelView(columns, dtypes, action_, on_action).key("merge")

    with v.Sheet() as main:
        v.Select(
            label="Transformation",
            v_model=op,
            items=list(action_mapping.values()),
            on_v_model=lambda v: on_action_type(action_mapping_reverse[v]),  # type: ignore
        )
        if op is not None:
            get_panel()

    return main



from playwright.sync_api import Page


class TransformHelper:
    def __init__(self, page: Page) -> None:
        self.page = page

    @property
    def menu_item(self):
        return self.page.locator('div[role="listbox"] >> text=Transformations')

    @property
    def dialog(self):
        return self.page.locator(".v-dialog__content--active")

    @property
    def add_transformation(self):
        return self.page.locator("text=Add transformation >> xpath=..")

    @property
    def insert_code(self):
        return self.dialog.locator('button:has-text("Run")')

    @property
    def preview(self):
        return self.page.locator('button:has-text("Preview")')

    def choose_dataframe(self, name):
        self.page.locator('div[role="button"]:has-text("DataFrame")').click()
        self.page.locator(f'div[role="option"] >> text={name}').click()

    def table_cell(self, row, column):
        return self.page.locator(f"tr:nth-child({row}) td:nth-child({column})")

    @property
    def filter_like(self):
        return TransformFilterLike(self.page, self)


class TransformFilterLike:
    def __init__(self, page: Page, transform: TransformHelper) -> None:
        self.page = page
        self.transform = transform

    @property
    def dialog(self):
        # it seems like vue=v-dialog does not work
        return self.page.locator("_vue=v-card >> _vue=v-card-title >> text=Filter like >> .. >> ..")

    @property
    def apply(self):
        return self.dialog.locator('button:has-text("Apply")')

    @property
    def column(self):
        return self.dialog.locator('_vue=v-select[label="Column"]')

    @property
    def column_text(self):
        return self.column.locator(".v-select__selection").text_content()

    @property
    def operator(self):
        return self.dialog.locator('_vue=v-select[label="Operator"]')

    @property
    def operator_text(self):
        return self.operator.locator(".v-select__selection").text_content()

    @property
    def value(self):
        return self.dialog.locator('_vue=v-text-field[label="Value"]')

    @property
    def value_text(self):
        return self.value.locator("input").input_value()

    @value_text.setter
    def value_text(self, value):
        self.value.locator("input").fill(value)

    @property
    def new_dataframe_name(self):
        return self.dialog.locator('_vue=v-text-field[label="New dataframe name"]')

    @property
    def new_dataframe_name_text(self):
        return self.new_dataframe_name.locator("input").input_value()

    @new_dataframe_name_text.setter
    def new_dataframe_name_text(self, value):
        self.new_dataframe_name.locator("input").fill(value)



import os
import pathlib
import site

from domino_code_assist.domino_api import get_domino_api
from domino_code_assist.settings import settings


def in_user_install_mode():
    return __file__.startswith(site.getuserbase())


def in_pre_run_script():
    if os.path.exists("/domino/launch/preRunScript.sh"):
        with open("/domino/launch/preRunScript.sh") as f:
            content = f.read()
            if " low-code-assistant" in content or "low_code_assistant" in content or " domino-code-assist" in content or "domino_code_assist" in content:
                return True
    return False


def symlink_solara_assets():
    prefix = site.getuserbase()

    src = prefix + "/share/solara/cdn"
    dst = pathlib.Path(prefix + "/share/jupyter/nbextensions/_solara")
    dst.mkdir(exist_ok=True, parents=True)
    dst_cdn = dst / "cdn"
    if not dst_cdn.exists():
        os.symlink(src, dst_cdn)


def write_requirements_txt():
    file = pathlib.Path("requirements.txt")
    content = "domino-code-assist\n"
    if file.exists():
        old_content = file.read_text()
        if not (
            "low_code_assistant" in old_content
            or "low-code-assistant" in old_content
            or "domino_code_assist" in old_content
            or "domino-code-assist" in old_content
        ):
            file.write_text(old_content + "\n" + content)
            return True
    else:
        file.write_text(content)
        return True
    return False


if in_user_install_mode():
    symlink_solara_assets()
    try:
        if not settings.domino_is_git_based:
            if not in_pre_run_script():
                if write_requirements_txt():
                    get_domino_api().sync_files("requirements.txt added by domino_code_assist")
    except Exception:
        pass



# import reacton
import sys

import reacton.ipyvuetify as v
import solara
from solara.components.file_browser import FileBrowser
from solara.components.file_drop import FileDrop
from solara.lab.toestand import Ref

from ..assistant import mixpanel
from ..domino_api import get_domino_api
from .data_source import DataSourcePanel
from .quick_start import QuickStart
from .state import load_data

supported_extensions = [
    "csv",
    "excel",
    "feather",
    "fwf",
    "gbq",
    "hdf",
    "html",
    "json",
    "orc",
    "parquet",
    "pickle",
    "sas",
    "spss",
    "sql",
    "sql_query",
    "sql_table",
    "stata",
    "table",
    "xml",
    "nocode",
]


@solara.component
def LoadDataPanel():
    progress, set_progress = solara.use_state(0, key="progress")
    load_data.use()
    tab = Ref(load_data.fields.tab).get()

    def on_tab():
        tabs = ["'Data Sources'", "'Datasets'", "'Project Files'", "'Uploads'", "'Demo'"]
        mixpanel.api.track_with_defaults(
            "interaction",
            {
                "section": "load_data",
                "type": "select tab",
                "tab": tabs[tab],
            },
        )

    solara.use_memo(on_tab, [tab])

    with solara.Div(style_="max-width: 700px; max-width: 700px") as main:
        with v.Tabs(v_model=tab, on_v_model=Ref(load_data.fields.tab).set):
            v.Tab(children=["Data Sources"])
            v.Tab(children=["Datasets"])
            v.Tab(children=["Project Files"])
            v.Tab(children=["Upload"])
            v.Tab(children=["Quick-start"])
            with v.TabItem():
                version = get_domino_api().get_domino_version()
                if not (version.startswith("5") or version.startswith("0")):
                    solara.Div(children=["Data Sources are not available in Domino 4.x and earlier."], class_="ma-4")
                elif sys.version_info.minor == 6:
                    solara.Div(children=["Data Sources are not available in Python 3.6."], class_="ma-4")
                else:
                    DataSourcePanel(
                        columns=None,
                        visible=tab == 0,
                    ).meta(ref="datasource")
            with v.TabItem():
                dataset_dir = Ref(load_data.fields.dataset_dir).get()
                if dataset_dir.exists():
                    with solara.Div(style_="height: calc(100vh - 166px); padding-bottom: 0px; overflow: hidden; display: flex; flex-direction: column;"):
                        solara.Div(class_="my-4", children=[f"Select a file with one of the following extensions: {', '.join(supported_extensions)}"])
                        with solara.Div(style_="display: flex; flex-direction: column; flex-grow: 1; overflow: hidden;"):
                            with solara.Div(style_="flex-grow: 1; overflow: hidden;"):
                                FileBrowser(
                                    directory=dataset_dir,
                                    on_directory_change=load_data.on_dataset_dir_change,
                                    on_file_name=Ref(load_data.fields.dataset_file).set,
                                ).key("DatasetsFileBrowser").meta(ref="datasets")
                            v.Switch(label="reactive").connect(Ref(load_data.fields.reactive))
                else:
                    solara.Div(children=["This project has no datasets."], class_="ma-4")
            with v.TabItem():
                with solara.Div(
                    style_="height: calc(100vh - 166px); overflow: auto; padding-bottom: 0px; overflow: hidden; display: flex; flex-direction: column;"
                ):
                    solara.Div(class_="my-4", children=[f"Select a file with one of the following extensions: {', '.join(supported_extensions)}"])
                    with solara.Div(style_="display: flex; flex-direction: column; flex-grow: 1; overflow: hidden;"):
                        FileBrowser(
                            directory=load_data.value.project_dir,
                            on_directory_change=load_data.on_project_dir_change,
                            on_file_name=Ref(load_data.fields.project_file).set,
                        ).key("ProjectFileBrowser").meta(ref="project")
                        v.Switch(label="reactive").connect(Ref(load_data.fields.reactive))
            with v.TabItem():
                solara.Div(class_="my-4", children=[f"Drag a file with one of the following extensions: {', '.join(supported_extensions)}"])
                FileDrop(
                    label="Drop dataset here",
                    on_total_progress=set_progress,
                    on_file=Ref(load_data.fields.upload_file).set,
                ).meta(ref="upload")
                v.ProgressLinear(value=progress)
                v.Switch(label="reactive").connect(Ref(load_data.fields.reactive))
            with v.TabItem():
                QuickStart()
                v.Switch(label="reactive").connect(Ref(load_data.fields.reactive))

    return main



import os
from pathlib import Path

import reacton
import reacton.ipyvuetify as v
import solara as sol

from domino_code_assist.assistant import drawer, mixpanel

from . import snippets, snippets_ui
from .snippets import EDIT_SNIPPET_PREFIX, Snippet


@reacton.component
def SnippetDrawer(set_code, open, on_close, save_count, edit_mode, on_edit_mode, writable_paths, overwrite_warning=None):
    root = {"label": "home", "id": 0, "leaf": False}
    path_, set_path = reacton.use_state([root])
    refresh_count, set_refresh_count = reacton.use_state(0)

    nr_of_snippets, get_items, get_content, adjust_path = reacton.use_memo(snippets.make_fns, [save_count, refresh_count])

    path = reacton.use_memo(lambda: adjust_path(path_), [adjust_path, path_])

    def on_apply():
        mixpanel.api.track_with_defaults(
            "inserted code",
            {
                "section": "Insert snippet",
            },
        )
        set_code(
            {
                "code": snippet_code,
                "meta": None,
            }
        )
        on_close()

    snippet_code, set_snippet_code = reacton.use_state("")

    def on_edit(content: Snippet):
        set_code(
            {
                "code": f"{EDIT_SNIPPET_PREFIX} {content.base_path / content.path / content.name}\n" + content.code,
            }
        )
        on_close()

    def on_delete(content: Snippet):
        os.remove(content.base_path / content.path / content.name)
        set_refresh_count(refresh_count + 1)

    def on_add_snippet(base_path, path, name):
        set_code(
            {
                "code": f"{EDIT_SNIPPET_PREFIX} {Path(base_path) / 'snippets' / Path(*[p['id'] for p in path[1:] if not p['leaf']] + [name + '.py'])}\n",
            }
        )
        on_close()

    with drawer.RightDrawer(
        open=open,
        on_open=lambda v: on_close() if not v else None,
        title="Snippets",
        edit=False,
        on_apply=on_apply,
        apply_disabled=not snippet_code,
        show_default_buttons=bool(nr_of_snippets),
        warning_widget=overwrite_warning,
    ) as main:
        with sol.Div(style_="min-width: 1200px; max-width: 1200px; height: 100%;"):
            if open:
                if nr_of_snippets:
                    snippets_ui.SnippetsPanel(
                        get_items,
                        get_content,
                        path,
                        set_path,
                        edit_mode,
                        on_edit_mode,
                        writable_paths,
                        on_edit,
                        on_delete,
                        on_add_snippet,
                        on_snippet=lambda snippet: set_snippet_code(snippet.code if snippet else ""),
                        on_refresh_count=set_refresh_count,
                    ).key("snippets")
                else:
                    with sol.Div(class_="text-center"):
                        v.Html(
                            tag="h3",
                            class_="ma-4",
                            children=["No snippets found. Snippets can be added by adding a shared project or shared dataset that has a snippets directory."],
                        )

    return main



from playwright.sync_api import Page


def mouse_move_middle(page: Page, locator):
    page.wait_for_timeout(100)  # seems less flakey if we wait 100msec
    box = locator.bounding_box()
    x, y = box["x"] + box["width"] / 2, box["y"] + box["height"] / 2
    # move 100 pixels above the cell...
    page.mouse.move(x, box["y"] - 100)
    page.wait_for_timeout(100)
    # and move to the middle of the cell, in 10 steps
    page.mouse.move(x, y, steps=10)



from jupyter_server.utils import url_path_join

from domino_code_assist.jupyter.mixpanel_handler import MixpanelHandler


def _jupyter_server_extension_paths():
    return [{"module": "domino_code_assist.jupyter.server_extension"}]


def _load_jupyter_server_extension(server_app):
    web_app = server_app.web_app

    host_pattern = ".*$"
    base_url = url_path_join(web_app.settings["base_url"])

    web_app.add_handlers(
        host_pattern,
        [
            (url_path_join(base_url, "/mixpanel/(.*)"), MixpanelHandler, {}),
        ],
    )


# For backward compatibility
load_jupyter_server_extension = _load_jupyter_server_extension



from functools import partial
from itertools import groupby
from typing import List

import plotly
import reacton
import reacton.ipyvuetify as v
import solara as sol
import solara.express
from pandas.core.frame import DataFrame
from solara.lab.components.cross_filter import (
    CrossFilterReport,
    CrossFilterSelect,
    CrossFilterSlider,
)

import domino_code_assist.deploy
from domino_code_assist import app, css, util
from domino_code_assist.thumbnail import Thumbnail


def to_widget(x):
    if isinstance(x, plotly.graph_objs._figure.Figure):
        return sol.FigurePlotly(x)
    else:
        # todo: add special cases as needed
        return x


def _make_grid_panel_builder(layout_item, markdown_cells):

    children = []
    if layout_item.get("i"):
        i = layout_item.get("i")
        if util.nb_locals.get(i) is not None:
            obj = util.nb_locals[i]
            if isinstance(obj, plotly.graph_objs._figure.Figure):
                children = [sol.FigurePlotly(obj)]
            elif isinstance(obj, DataFrame):
                children = [sol.CrossFilterDataFrame(obj, items_per_page=10, scrollable=True)]
            elif isinstance(obj, app.ReactiveDf):
                children = [app.DropDataframe(obj.var_name, obj.label)]
            else:
                children = [obj]
        elif markdown_cells and i in markdown_cells:
            children = [sol.Markdown(markdown_cells[i])]

    with v.Card(style_="height: 100%;") as panel:
        v.CardText(style_="height: 100%; max-height: 100%; overflow: auto;", children=children)

    return panel


@reacton.component
def CrossFilterScope(children: List[reacton.core.Element] = []):
    sol.provide_cross_filter()

    return sol.Div(children=children)


@reacton.component
def CardGridLayoutBuilder(obj_vars, layout, on_layout, markdown_cells, on_toggle):
    objects_on = [k["i"] for k in layout]

    group_names = ["Plots", "Markdown", "Widgets", "Dataframes", "Other"]

    def group_obj_vars():
        def sort_vars(var):
            obj = util.nb_locals.get(var)
            if obj is None:
                # markdown
                return 1
            if isinstance(obj, plotly.graph_objs._figure.Figure) or (
                isinstance(obj, reacton.core.Element) and obj.component is solara.express.CrossFilteredFigurePlotly
            ):
                return 0
            if isinstance(obj, reacton.core.Element) and obj.component in [CrossFilterSlider, CrossFilterSelect, CrossFilterReport]:
                return 2
            if isinstance(obj, DataFrame) or (isinstance(obj, reacton.core.Element) and obj.component is DataTableCrossFiltered):
                return 3
            return 4

        grouped = groupby(
            sorted(obj_vars + list((markdown_cells or {}).keys()), key=sort_vars),
            sort_vars,
        )
        nb_order = list(util.nb_locals.keys())
        return [
            (group_names[group], sorted(rows, key=lambda x: nb_order.index(x) if x in nb_order else list(markdown_cells.keys()).index(x)))
            for (group, rows) in grouped
        ]

    grouped_obj_vars = reacton.use_memo(group_obj_vars, [obj_vars, markdown_cells])

    def toggle_layout(on, id):
        on_toggle(on)

        def update(layout):
            layout_contains_id = any([k for k in layout if k["i"] == id])

            if on:
                top_y = max([k["y"] + k["h"] for k in layout]) if layout else 0
                return layout + [dict(i=id, w=6, h=5, x=0, y=top_y)] if not layout_contains_id else layout
            else:
                return [k for k in layout if k["i"] != id] if layout_contains_id else layout

        on_layout(lambda state: update(state))

    items = {layout_item["i"]: _make_grid_panel_builder(layout_item, markdown_cells) for layout_item in layout}

    def preselect():
        if not layout and "plot_quick_start" in obj_vars:
            on_layout(
                lambda state: [
                    *([{"i": "df_quick_start", "w": 6, "h": 10, "x": 6, "y": 0}] if "df_quick_start" in obj_vars else []),
                    {"i": "plot_quick_start", "w": 6, "h": 10, "x": 0, "y": 0},
                ]
            )
        elif not layout and grouped_obj_vars and grouped_obj_vars[0][0] == "Plots":
            on_layout(lambda state: [{"i": grouped_obj_vars[0][1][0], "w": 6, "h": 10, "x": 0, "y": 0}])

    reacton.use_memo(preselect, [])

    with v.Sheet(class_="domino-plotly-auto-height domino-card-grid", style_="height: 100%; max-height: 100%; overflow: hidden; margin-left: -26px;") as main:
        css.Css()
        with sol.Div(style_="display: flex; height: 100%; max-height: 100%;"):
            with sol.Div(style_="width: 256px; overflow: auto; padding-left: 26px; background-color: var(--jp-border-color2);"):
                with sol.Div(style_="width: 200px"):
                    for group, group_vars in grouped_obj_vars:
                        sol.Div(style_="margin-top: 16px; margin-bottom: -8px; font-weight: bold", children=[group + ": "])
                        for obj_var in group_vars:
                            label = obj_var
                            if markdown_cells and obj_var in markdown_cells:
                                label = markdown_cells[obj_var].replace("\n", " ")
                                length = len(label)
                                label = label[:28] + ("..." if length > 28 else "")
                            with sol.Div():
                                with v.Sheet(elevation=1, class_="px-1"):
                                    v.Switch(label=label, v_model=obj_var in objects_on, on_v_model=partial(toggle_layout, id=obj_var))
                                    with CrossFilterScope():
                                        with Thumbnail():
                                            _make_grid_panel_builder({"i": obj_var}, markdown_cells)

            with sol.Div(style_="flex-grow: 1; overflow: auto"):
                with CrossFilterScope():
                    sol.GridLayout(items=items, grid_layout=layout, resizable=True, draggable=True, on_grid_layout=on_layout)

    return main


# for backward compatibility
Preview = domino_code_assist.deploy.Deployer
CardGridLayout = domino_code_assist.CardGridLayout
DataTableCrossFiltered = solara.CrossFilterDataFrame



import json

import domino_code_assist.actions_store as actions_store
import domino_code_assist.viz.state

from . import action, crossfilter_widgets
from .settings import settings


class AEncoder(json.JSONEncoder):
    def default(self, obj):
        encoded = action.to_json(obj)
        if not encoded:
            encoded = actions_store.to_json(obj)
        if not encoded:
            encoded = crossfilter_widgets.to_json(obj)
        if not encoded:
            encoded = domino_code_assist.viz.state.to_json(obj)
        if not encoded:
            encoded = domino_code_assist.automl.store.to_json(obj)
        return encoded if encoded else json.JSONEncoder.default(self, obj)


def from_json(dct):
    decoded = action.from_json(dct)
    if not decoded:
        decoded = actions_store.from_json(dct)
    if not decoded:
        decoded = domino_code_assist.viz.state.from_json(dct)
    if not decoded:
        decoded = crossfilter_widgets.from_json(dct)
    if not decoded:
        decoded = domino_code_assist.automl.store.from_json(dct)
    return decoded if decoded else dct


def dumps(value):
    return json.dumps(value, cls=AEncoder)


def loads(str):
    try:
        data = json.loads(str, object_hook=from_json)
        # convert previously stored action list to action state
        if type(data) == list and len(data) > 0 and isinstance(data[0], action.Action):
            data = actions_store.ActionsState(preview=True, actions=data)
        return data
    except Exception as e:
        if settings.domino_code_assist_dev:
            raise e



from dataclasses import replace
from pathlib import Path
from typing import cast

import reacton
from solara.lab import Ref

from domino_code_assist import action
from domino_code_assist.assistant import drawer, mixpanel
from domino_code_assist.settings import settings

from .. import util
from ..serialize import dumps
from . import state
from .panel import LoadDataPanel


@reacton.component
def LoadDataDrawer(edit_action, set_code, open, edit, on_close, overwrite_warning=None):
    state.load_data.use()

    def reset_data():
        # reset state
        state.load_data.set(state.LoadData())

    reacton.use_memo(reset_data, [open])
    load_data_value = state.load_data.use_value()
    tab = load_data_value.tab

    def set_action(new_action):
        data_type = None
        data_source_type = None
        if isinstance(new_action, action.ActionOpen):
            if (new_action.filename or "").startswith(settings.domino_datasets_dir):
                data_type = "dataset"
            else:
                data_type = "file"
        elif isinstance(new_action, action.ActionDownloadDataSource):
            data_type = "data source"
            data_source_type = new_action.type_ or ""
        elif isinstance(new_action, action.ActionDemo):
            data_type = "demo"
        mixpanel.api.track_with_defaults(
            "inserted code",
            {
                "section": "load_data",
                "data_type": data_type,
                **({"data_source_type": data_source_type} if data_source_type else {}),
            },
        )
        set_code({"code": new_action.render_code("df") + f"\n{new_action.df_var_out}", "meta": dumps(new_action)})
        on_close()

    def on_file_name(filename: str, df_var: str, reactive: bool):
        set_action(action.ActionOpen(filename=filename, df_var_out=df_var, reactive=reactive))

    def on_apply():
        if tab == 0:
            action_ds = Ref(state.load_data.fields.data_source).get()
            action_ds = replace(action_ds, df_var_out=Ref(state.load_data.fields.df_out).value)
            set_action(action_ds)
        elif tab == 1 and load_data_value.dataset_file:
            on_file_name(load_data_value.dataset_file, load_data_value.df_out, load_data_value.reactive)
        elif tab == 2 and load_data_value.project_file:
            on_file_name(load_data_value.project_file, load_data_value.df_out, load_data_value.reactive)
        elif tab == 3 and load_data_value.upload_file:
            dest_name = util.copy_upload(load_data_value.upload_file["file_obj"], load_data_value.upload_file["name"])
            on_file_name(dest_name, load_data_value.df_out, load_data_value.reactive)
        elif tab == 4:
            set_action(action.ActionDemo(df_name=load_data_value.demo_df_name, df_var_out=load_data_value.df_out, reactive=load_data_value.reactive))

    def set_edit():
        if open and edit:
            if isinstance(edit_action, action.ActionOpen):
                filename = cast(action.ActionOpen, edit_action).filename
                if filename is None:
                    state.load_data.update(tab=0, df_out=edit_action.df_var_out)
                elif "domino_code_assist_data" in filename:
                    state.load_data.update(tab=3, df_out=edit_action.df_var_out)
                elif filename.startswith(settings.domino_datasets_dir):
                    state.load_data.update(tab=1, dataset_dir=Path(filename or "mypy").parent, df_out=edit_action.df_var_out, reactive=edit_action.reactive)
                else:
                    state.load_data.update(tab=2, project_dir=Path(filename).parent, df_out=edit_action.df_var_out, reactive=edit_action.reactive)
            elif isinstance(edit_action, action.ActionDownloadDataSource):
                state.load_data.update(tab=0, data_source=edit_action, df_out=edit_action.df_var_out)
            elif isinstance(edit_action, action.ActionDemo):
                state.load_data.update(tab=4, demo_df_name=edit_action.df_name, df_out=edit_action.df_var_out)

    reacton.use_memo(set_edit, [open, edit, edit_action])

    with drawer.RightDrawer(
        open=open,
        on_open=lambda v: on_close() if not v else None,
        title="Load data",
        edit=bool(edit),
        apply_disabled=not state.load_data.get().valid(),
        on_apply=on_apply,
        show_var_out=True,
        var_out=state.load_data.get().df_out,
        on_var_out=Ref(state.load_data.fields.df_out).set,
        warning_widget=overwrite_warning,
    ) as main:
        if open:
            LoadDataPanel()

    return main



import json
import logging
import os
import threading
import time
import uuid
from abc import ABC, abstractmethod
from typing import Dict, Optional
from urllib.parse import quote

import ipyvuetify as vy
import requests
import traitlets

from domino_code_assist import __version__
from domino_code_assist.domino_api import get_domino_api
from domino_code_assist.settings import settings

logger = logging.getLogger("domino_code_assist")

user_info: Dict[str, str] = {}


class IMixpanel(ABC):
    @abstractmethod
    def track(self, event: str, props: Optional[Dict] = None):
        pass

    def identify(self, unique_id: str):
        pass

    def track_with_defaults(self, event: str, properties: Optional[Dict] = None):
        props: Dict = {
            "LCA_version": __version__,
            "Domino_version": get_domino_api().get_domino_version(),
            "LCA_language": "python",
            **user_info,
            **(properties or {}),
        }
        self.track(f"LCA {event}", props)


class NopMixpanel(IMixpanel):
    def track(self, event: str, props: Optional[Dict] = None):
        pass

    def identify(self, unique_id: str):
        pass


class MixpanelWidget(vy.VuetifyTemplate):
    template_file = (__file__, "mixpanel.vue")

    project_token = traitlets.Unicode(allow_none=True).tag(sync=True)
    debug = traitlets.Bool(False).tag(sync=True)
    initialized = traitlets.Bool(False).tag(sync=True)
    unreachable_in_frontend = traitlets.Bool(False).tag(sync=True)
    frontend_props = traitlets.Dict().tag(sync=True)
    mixpanel_enabled = traitlets.Bool(False, allow_none=True).tag(sync=True)

    def track(self, event: str, props: Optional[Dict] = None):
        self.send({"method": "track", "args": [event, props]})

    def identify(self, unique_id: str):
        self.send({"method": "identify", "args": [unique_id]})


class FrontendMixpanel(IMixpanel):
    def __init__(self, w: MixpanelWidget):
        self.w = w

    def track(self, event: str, props: Optional[Dict] = None):
        self.w.track(event, props)

    def identify(self, unique_id: str):
        self.w.identify(unique_id)


class BackendMixpanel(IMixpanel):
    def __init__(self, project_token: str, frontend_props: Dict):
        self.project_token = project_token
        self.frontend_props = frontend_props
        self.unique_id = str(uuid.uuid4())
        self.user_id = None
        self.fail_count = 0
        self.lock = threading.Lock()

        try:
            info = get_domino_api().get_user_info()
            self.user_id = info["id"] if info else None
            if info:
                global user_info
                user_info = {
                    "username": info["userName"],
                    "email": info["email"],
                }

        except Exception as e:
            if settings.domino_code_assist_dev:
                logger.info(e)

    def track(self, event: str, props: Optional[Dict] = None):

        event_item = {
            "event": event,
            "properties": {
                "token": self.project_token,
                "time": int(round(time.time() * 1000)),
                "distinct_id": self.user_id or self.unique_id,
                "$insert_id": str(uuid.uuid4()),
                **(self.frontend_props or {}),
                **({"$user_id": self.user_id} if self.user_id else {}),
                **(props or {}),
            },
        }

        def run():
            try:
                requests.post(
                    "https://api.mixpanel.com/track/",
                    headers={"content-type": "application/x-www-form-urlencoded"},
                    data=f"data={quote(json.dumps([event_item]))}",
                    timeout=1,
                )
                with self.lock:
                    self.fail_count = 0
            except Exception as e:
                with self.lock:
                    self.fail_count += 1
                if settings.domino_code_assist_dev:
                    logger.info(e)

        with self.lock:
            if self.fail_count < 3:
                threading.Thread(target=run).start()

    def identify(self, unique_id: str):
        pass


mixpanel_widget = None
server_software = os.environ.get("SERVER_SOFTWARE", "")

if not (server_software and server_software.startswith("solara")):
    mixpanel_widget = MixpanelWidget(project_token="a442c4dbf420b59cdac95cfb07f57cbb", debug=False)

api: IMixpanel = FrontendMixpanel(mixpanel_widget) if mixpanel_widget else NopMixpanel()

if mixpanel_widget:

    def init(change):
        if change["new"]:
            info = get_domino_api().get_user_info()
            if info:
                unique_id = info["id"] if info else None
                if unique_id:
                    api.identify(unique_id)
                global user_info
                user_info = {
                    "username": info["userName"],
                    "email": info["email"],
                }
            api.track_with_defaults("initialized")

    mixpanel_widget.observe(init, names="initialized")

    def unreachable(change):
        if change["new"]:
            global api
            if mixpanel_widget:
                api = BackendMixpanel(mixpanel_widget.project_token, mixpanel_widget.frontend_props) if mixpanel_widget.mixpanel_enabled else NopMixpanel()
                api.track_with_defaults("initialized")

    mixpanel_widget.observe(unreachable, names="unreachable_in_frontend")



import textwrap

import humanize
import pandas as pd
import reacton
import reacton.ipyvuetify as v
import solara
from solara.lab import Ref

from ..domino_api import get_domino_api
from .state import ReactiveDataSource, load_data

supported_types = ["RedshiftConfig", "SnowflakeConfig"]


query_snowflake_meta = """select
                    TABLE_CATALOG as "database_name",
                    TABLE_SCHEMA as "schema_name",
                    TABLE_NAME as "name",
                    IFNULL(ROW_COUNT, -1) as "rows"
                    from INFORMATION_SCHEMA.TABLES
                    where TABLE_SCHEMA != 'INFORMATION_SCHEMA'"""


@reacton.component
def Panel(ds):
    data_source: ReactiveDataSource = load_data.data_source
    data_source.use()

    def get_tables(_cancel):
        if not ds:
            return None
        if ds.datasource_type == "SnowflakeConfig":
            if ds.config.get("database"):
                return ds.query(query_snowflake_meta).to_pandas()
            else:
                df_tables = ds.query("show tables").to_pandas()
                views = ds.query("show views").to_pandas()
                views = views.loc[views["schema_name"] != "INFORMATION_SCHEMA"]
                views = views.loc[views["database_name"] != "SNOWFLAKE"]
                views["rows"] = -1
                views["bytes"] = -1
                return pd.concat([df_tables, views])
        if ds.datasource_type == "RedshiftConfig":
            res = ds.query(
                textwrap.dedent(
                    """\
                select tab.table_schema as schema_name,
                       tab.table_name as name,
                       class.reltuples as rows
                from svv_tables tab
                         join pg_class class
                              on tab.table_name = class.relname
                where tab.table_type = 'BASE TABLE'
                  and tab.table_schema not in('pg_catalog','information_schema')
                  and class.reltuples > 1
                """
                )
            )
            return res.to_pandas()
        raise RuntimeError(f"Unknown datasource: {ds.datasource_type}")

    tables_result: solara.Result = solara.hooks.use_thread(get_tables, [ds])

    tables = tables_result.value

    database_items = []
    schema_items = []
    table_items = []
    names = None
    rows = None

    if tables is not None and ds is not None:
        database_items = tables["database_name"].unique().tolist() if "database_name" in tables else [ds.config.get("database")]
        database = data_source.value.database
        schema = data_source.value.schema
        if database:
            if ds.datasource_type == "SnowflakeConfig":
                schema_items = tables.loc[tables["database_name"] == database]["schema_name"].unique().tolist()
            else:
                schema_items = tables["schema_name"].unique().tolist()
            if schema:
                if ds.datasource_type == "SnowflakeConfig":
                    table_rows = tables[(tables["database_name"] == database) & (tables["schema_name"] == schema)]
                else:
                    table_rows = tables.loc[tables["schema_name"] == schema]
                names = table_rows["name"].tolist()
                rows = table_rows["rows"].tolist()

                def row_str(r):
                    return f"({humanize.intword(r)} rows)" if r >= 0 else " (view)"

                table_items = [{"value": n, "text": f"{n} {row_str(r)}"} for n, r in zip(names, rows)]

    def get_schema_spec(_cancel):
        database = data_source.value.database
        schema = data_source.value.schema
        if not (ds and database and schema):
            return None
        elif ds.datasource_type == "RedshiftConfig":
            return (
                ds.query(
                    textwrap.dedent(
                        f"""\
                    select table_name, column_name
                    from information_schema.columns
                    where table_schema = '{schema}'
                    and table_catalog = '{database}'
                    order by table_name, ordinal_position
                    """
                    )
                )
                .to_pandas()
                .groupby("table_name")["column_name"]
                .apply(list)
                .to_dict()
            )
        elif ds.datasource_type == "SnowflakeConfig":
            return ds.query(f"SHOW COLUMNS IN SCHEMA {database}.{schema}").to_pandas().groupby("table_name")["column_name"].apply(list).to_dict()

    spec_result: solara.Result = solara.hooks.use_thread(get_schema_spec, [ds and ds.datasource_type, data_source.value.database, data_source.value.schema])

    def pre_select_database():
        if len(database_items) == 1 and database_items[0] != data_source.value.database:
            data_source.set_database(database_items[0])

    solara.use_memo(pre_select_database, [database_items])

    def pre_selelect_schema():
        if len(schema_items) == 1 and schema_items[0] != data_source.value.schema:
            data_source.set_schema(schema_items[0])

    solara.use_memo(pre_selelect_schema, [schema_items])

    def check_sample():
        if data_source.value.use_query:
            data_source.set_sample(False)
            return

        if tables is None:
            return

        name_index = not data_source.value.use_query and names and data_source.value.table and names.index(data_source.value.table)
        nr_rows = name_index is not None and rows and rows[name_index]

        data_source.set_sample(nr_rows and nr_rows > 3000)

    solara.use_memo(check_sample, [data_source.value.table, data_source.value.use_query])

    with v.Sheet() as main:
        if tables_result.error:
            error = tables_result.error
            error_msg = f"{type(error).__name__}: {str(error)}"
            if isinstance(error, KeyError):
                error_msg = (
                    error_msg
                    + ". This may be due to an environment with an incompatible version of the Domino-API-client (v"
                    + f"{get_domino_api().get_client_version()}) for this Domino version (v{get_domino_api().get_domino_version()})."
                )
            solara.Warning(error_msg)
        elif not ds or tables_result.state == solara.ResultState.RUNNING:
            v.ProgressLinear(indeterminate=True).meta(ref="tables_loading")
        else:
            with v.Row():
                with v.Col():
                    v.Select(label="Database", items=database_items, v_model=data_source.value.database, on_v_model=data_source.set_database).meta(
                        ref="database"
                    )
                with v.Col():
                    v.Select(label="Schema", items=schema_items, v_model=data_source.value.schema, on_v_model=data_source.set_schema).meta(ref="schema")
            with v.Row():
                with v.Col():
                    v.Switch(label="Use query").connect(Ref(data_source.fields.use_query)).meta(ref="use_query")
            with v.Row():
                if data_source.value.use_query:
                    with v.Col():
                        solara.SqlCode(query=data_source.value.query or "", tables=spec_result.value or {}, on_query=Ref(data_source.fields.query).set).meta(
                            ref="query_editor"
                        )
                else:
                    with v.Col():
                        v.Select(label="Table", items=table_items).connect(Ref(data_source.fields.table))

    return main



from typing import Callable, Dict, Optional

import solara.lab

save_notebook: Optional[Callable[[Callable[[bool], None]], None]] = None
markdown_cells: Optional[Dict] = None
notebook_browser_path = solara.lab.Reactive("")



import ipyvuetify as v
import ipywidgets
import numpy as np
import pandas as pd
import plotly.express as px
import solara
import traitlets
import vaex
from pandas.api.types import is_numeric_dtype


@solara.memoize(key=lambda df, column: id(df[column]) if column in df.columns else None)
def get_column_description(df, column):
    if column not in df.columns:
        return None, None

    column = column
    df_vx = vaex.from_pandas(df[[column]])
    unique_future = df_vx[column].unique(limit=1000, dropna=True, delay=True)
    description = df_vx.describe().to_dict()[column]
    value_counts = None

    if unique_future.value:
        value_counts = df_vx[column].value_counts()

    for k in ["count", "mean", "std", "min", "max", "NA"]:
        if isinstance(description.get(k), np.ndarray):
            description[k] = description[k].tolist()
    description["data_type"] = repr(description["data_type"])

    description["unique"] = unique_future.value
    try:
        description["unique"].sort()
    except Exception:
        pass

    dtype = str(df[column].dtype)
    if description["data_type"] == dtype:
        description.pop("data_type")

    nbins = 20
    histogram_fig = None
    if value_counts is not None and (len(value_counts) <= nbins or not is_numeric_dtype(df[column])):
        histogram_fig = px.bar(x=value_counts.index, y=value_counts.values)
    else:
        if is_numeric_dtype(df[column]):
            binby = df_vx.count(binby=df_vx[column], limits=[description["min"], description["max"] + 0.000000000001], shape=nbins)
            bins = df_vx.bin_edges(None, limits=[description["min"], description["max"]], shape=nbins)[:-1]
            histogram_fig = px.bar(pd.DataFrame(data={"x": bins.tolist(), "y": binby.tolist()}), x="x", y="y")
    if histogram_fig:
        histogram_fig.update_layout(
            bargap=0,
            margin=dict(l=0, r=0, t=0, b=0),
            xaxis_title=None,
            yaxis={
                "showgrid": False,
                "zeroline": False,
                "visible": False,
            },
        )

    description = {
        "description": description,
        "dtype": dtype,
    }

    return description, histogram_fig


class ColumnDescriptionWidget(v.VuetifyTemplate):
    template_file = (__file__, "column_description.vue")

    column = traitlets.Unicode(allow_none=True).tag(sync=True)
    info = traitlets.Dict(allow_none=True).tag(sync=True)
    histogram = traitlets.Any(allow_none=True).tag(sync=True, **ipywidgets.widget_serialization)
    loading = traitlets.Bool(False).tag(sync=True)
    error = traitlets.Unicode(allow_none=True).tag(sync=True)


@solara.component
def ColumnDescription(df, column):
    last_column = solara.use_previous(column, column is not None)
    # the column we use for the description, is the current column, or when None, the last column
    # this makes sure we finish a calculation when the user hovers aways from the column info menu
    used_column = column or last_column
    stats_result: solara.Result = get_column_description.use_thread(df, used_column)
    loading = stats_result.state == solara.ResultState.RUNNING
    info, histogram_fig = stats_result.value if stats_result.value is not None else (None, None)

    def make_histogram():
        if histogram_fig:
            # force creation of a new figure widget when the plotly figure changes
            # if solara gets a FigurePlotly component that is better at updating
            # data we can avoid that
            return solara.FigurePlotly(histogram_fig).key(str(id(histogram_fig)))
        return None

    # When the histogram_fig becomes None or we change the histogram_fig we create a new
    # element, otherwise we reuse the previous one
    histogram = solara.use_memo(make_histogram, [bool(histogram_fig) and id(histogram_fig)])
    return ColumnDescriptionWidget.element(
        column=used_column,
        info=info,
        histogram=histogram,
        loading=loading,
        error=repr(stats_result.error) if stats_result.state == solara.ResultState.ERROR else None,
    )



import threading
from dataclasses import dataclass, field, replace
from enum import Enum
from typing import Any, List, Optional, Tuple

import reacton

from domino_code_assist.action import Action, ActionUseNbLocals, render_code_chunks
from domino_code_assist.assistant import mixpanel

e_insert_index = "insert_index"
e_edit_index = "edit_index"
e_delete_index = "delete_index"
e_insert_action = "insert_action"
e_edit_action = "edit_action"
e_undo = "undo"
e_redo = "redo"
e_reset = "reset"


@dataclass(frozen=True)
class DfWrapper:
    df: Any = field(repr=False, hash=None, compare=False, default_factory=lambda: None)
    actions: Optional[Tuple[Action]] = None

    @property
    def columns(self):
        if self.df is None:
            return None
        return tuple(self.df.columns.tolist()) if self.has_df else None

    @property
    def has_df(self):
        return self.df is not None


class ThreadState(Enum):
    RUNNING = 1
    CANCELLED = 2
    DONE = 3


@dataclass(frozen=True)
class ActionsState:
    df_wrapper: DfWrapper = DfWrapper()
    df_exec_status: ThreadState = ThreadState.DONE

    actions: List[Action] = field(default_factory=lambda: [])
    code_chunks: List[str] = field(default_factory=lambda: [])
    columns_for_action: List[List[str]] = field(default_factory=lambda: [])
    dtypes_for_action: List[List[str]] = field(default_factory=lambda: [])

    error: Optional[Tuple[int, str]] = None
    index_of_edit_action: Optional[int] = None
    index_of_insert_action: Optional[int] = None

    undo_stack: List[List[Action]] = field(default_factory=lambda: [])
    redo_stack: List[List[Action]] = field(default_factory=lambda: [])
    preview: bool = False


def actions_reducer(state: ActionsState, event) -> ActionsState:
    event_type, payload = event
    # print("reducer", state, event_type, payload)

    if event_type == e_insert_index:
        return replace(state, index_of_insert_action=payload)

    if event_type == e_edit_index:
        return replace(state, index_of_edit_action=payload)

    if event_type == e_insert_action:
        if isinstance(payload, Action):
            index = state.index_of_insert_action
        else:
            index = payload[0]
            payload = payload[1]

        if index is None:
            _actions = state.actions + [payload]
            mixpanel.api.track_with_defaults(
                "interaction",
                {
                    "section": "transformations",
                    "type": "add",
                    "transformation_type": type(payload).__name__,
                },
            )
        else:
            _actions = state.actions[: index + 1] + [payload] + state.actions[index + 1 :]
            mixpanel.api.track_with_defaults(
                "interaction",
                {
                    "section": "transformations",
                    "type": "insert",
                    "transformation_type": type(payload).__name__,
                },
            )
        return replace(state, undo_stack=state.undo_stack + [state.actions], redo_stack=[], actions=_actions, index_of_insert_action=None)

    if event_type == e_edit_action:
        index = state.index_of_edit_action
        if index is None:
            return state
        mixpanel.api.track_with_defaults(
            "interaction",
            {
                "section": "transformations",
                "type": "edit",
                "transformation_type": type(payload).__name__,
            },
        )
        return replace(
            state,
            undo_stack=state.undo_stack + [state.actions],
            redo_stack=[],
            actions=state.actions[:index] + [payload] + state.actions[index + 1 :],
            index_of_edit_action=None,
        )

    if event_type == e_delete_index:
        index = payload
        if index is None:
            return state
        mixpanel.api.track_with_defaults(
            "interaction",
            {"section": "transformations", "type": "delete", "transformation_type": type(state.actions[index]).__name__},
        )
        return replace(state, undo_stack=state.undo_stack + [state.actions], redo_stack=[], actions=state.actions[:index] + state.actions[index + 1 :])

    if event_type == e_undo:
        mixpanel.api.track_with_defaults(
            "interaction",
            {
                "section": "transformations",
                "type": "undo",
            },
        )
        return replace(state, actions=state.undo_stack[-1], undo_stack=state.undo_stack[:-1], redo_stack=state.redo_stack + [state.actions])

    if event_type == e_redo:
        item = state.redo_stack[-1]
        mixpanel.api.track_with_defaults(
            "interaction",
            {
                "section": "transformations",
                "type": "redo",
            },
        )
        return replace(state, actions=item, redo_stack=state.redo_stack[:-1], undo_stack=state.undo_stack + [state.actions])

    if event_type == e_reset:
        return ActionsState()

    return state


def use_execute_df(state: ActionsState, set_state, nb_locals=None):
    def execute():
        def run():
            chunks = render_code_chunks(state.actions)
            set_state(
                lambda state: replace(
                    state,
                    df_exec_status=ThreadState.RUNNING,
                    code_chunks=chunks,
                    error=None,
                )
            )
            error = None
            df = None
            try:
                columns_for_action = []
                dtypes_for_action = []
                scope = {**nb_locals} if nb_locals else {}
                if nb_locals and not state.preview:
                    scope = nb_locals
                actions = state.actions
                if state.actions and isinstance(state.actions[0], ActionUseNbLocals):
                    df = scope[actions[0].df_var_out]
                    actions = actions[1:]
                    columns_for_action = [df.columns.values.tolist()]
                    dtypes_for_action = [{k: v.name for k, v in df.dtypes.items()}]
                for index, (action, chunk) in enumerate(zip(actions, chunks)):
                    if chunk:
                        exec(chunk, scope)
                    df = scope[action.df_var_out]
                    columns_for_action.append(df.columns.values.tolist())
                    dtypes_for_action.append({k: v.name for k, v in df.dtypes.items()})
            except Exception as e:
                error = (index, str(e))

            set_state(
                lambda state: replace(
                    state,
                    df_wrapper=DfWrapper(df, actions=state.actions),
                    error=error,
                    columns_for_action=columns_for_action,
                    dtypes_for_action=dtypes_for_action,
                    df_exec_status=ThreadState.DONE,
                )
            )

        if not state.actions:
            set_state(lambda state: replace(state, df_wrapper=DfWrapper(), error=None, columns_for_action=[], code_chunks=[]))
            return

        threading.Thread(target=run).start()

    reacton.use_effect(execute, [state.actions])


def to_json(obj):
    if isinstance(obj, ActionsState):
        return {
            "type_action_state": {
                "actions": obj.actions,
                "undo_stack": obj.undo_stack,
                "redo_stack": obj.redo_stack,
                "preview": obj.preview,
            }
        }


def from_json(dct):
    if "type_action_state" in dct:
        dct = dct["type_action_state"]
        return ActionsState(**dct)



"""Dominocode"""

__version__ = "1.3.0"

import os

# isort: skip_file

import domino_code_assist.logging_workaround

from . import data
from .assistant import init
from .components import MarkdownFromCell, CardGridLayout
from .util import mlflow_log_notebook

try:
    if not os.environ.get("DOMINO_PROJECT_ID"):
        from dotenv import load_dotenv

        load_dotenv()
except ImportError:
    pass


def _prefix():
    import sys
    from pathlib import Path

    prefix = sys.prefix
    here = Path(__file__).parent
    # for when in dev mode
    if (here.parent / "prefix").exists():
        prefix = str(here.parent)
    return prefix


def _jupyter_labextension_paths():
    return [
        {
            "src": f"{_prefix()}/prefix/share/jupyter/labextensions/domino-code-assist/",
            "dest": "domino-code-assist",
        }
    ]


def _jupyter_nbextension_paths():
    return [
        {
            "section": "notebook",
            "src": f"{_prefix()}/prefix/share/jupyter/nbextensions/domino-code-assist/",
            "dest": "domino-code-assist",
            "require": "domino-code-assist/extension",
        }
    ]


def _jupyter_server_extension_points():
    return [
        {
            "module": "myextension.app",
        }
    ]



# flake8: noqa
md1 = """\
# Palmer penguins

[The Palmer penguins dataset](https://github.com/allisonhorst/palmerpenguins) is a great dataset for data exploration & visualization. 
"""

md2 = """\
## Load data
Code Assist includes this dataset, so you can start exploring Code Assist directly.
"""

md3 = """\
# Transformations: filter out rows

For our analysis, we like to get rid of some rows that have missing data. Using Code Assist, we can filter out these rows using a UI without having to remember the proper Pandas syntax.

You can edit the transformation by hovering above the next code cell, hover about the blue Code Assist button,  and choosing "Edit".
"""

md4 = """\
## Visualizations: exploring penguin bills

To explore the bills of the penguins in this dataset, we create a scatter plot of `bill_length` vs `bill_depth` using the Code Assist UI.

We split the dataset into the two sexes, and three species. Without Code Assist, we would probably have to consult the plotly documentation. Using Code Assist, the options are readily available to us.

We use the "crossfilter" option, which will apply selections made in this plot to be applied as a filter in all other plots and widgets.

You can edit the visualizations by hovering above the next code cell, hover about the blue Code Assist button, and choosing "Edit."
"""

md5 = """\
# Visualizations: exploring the islands

Do bill lengths depend on the island the penguins live on? To explore this question, we generate a second visualization, where we plot the average `bill_length` per island.

Because we again enabled the "crossfilter" options, we can select one or multiple island by dragging over the histograms. This selection will be applied as a filter in the scatter plot.


You can edit the visualizations by hovering above the next code cell, hover about the blue Code Assist button, and choosing "Edit."
"""

md6 = """\
# Crossfilter widgets: selecting the year

To explore any effect or change over the years, we add a widget that allows us to select one or multiple years. This selection is also applied as a filter to the other two visualizations.

You can edit the widget by hovering above the next code cell, hover about the blue Code Assist button, and choosing "Edit".
"""

md7 = """\
The following two markdown cells will be used in the app we will create.
"""

md8 = """\
## Usage

  * Drag in the scatter plot to select a group of penguins. This selection will be applied as a filter to the histogram.
  * Drag to select histograms to select islands. The island selection will be applied as a filter to the scatter plot.
  * Select years to filter the years in the histogram and scatter plot.
  * Double-click a plot to clear the filter.
"""

md9 = """\
## Questions

  * Is there an island that only hosts 1 species?
  * Is there a species that is present on all three islands?
"""

md10 = """\
# Creating a penguin app

We now combined the visualizations, the widget, and a few markdown cells into an interactive web app. We can lay out out the visual elements on a grid by dragging them around.

With a single click on "Deploy app", we can deploy our app, so we can share it with others.

You can edit the app by hovering above the next code cell, hover about the blue Code Assist button, and choosing "Edit".
"""



import ipyvuetify as v
import solara
import traitlets


class SelectDescriptionWidget(v.VuetifyTemplate):
    template_file = (__file__, "select_description.vue")

    label = traitlets.Unicode().tag(sync=True)
    items = traitlets.List(allow_none=True).tag(sync=True)
    value = traitlets.Any().tag(sync=True)


@solara.component()
def SelectDescription(label, items, value, on_value):
    return SelectDescriptionWidget.element(label=label, value=value, items=items, on_value=on_value)



from pathlib import Path
from typing import Callable, Dict, List, Optional, cast

import reacton
import reacton.ipyvuetify as v
import solara as sol
import traitlets
from ipyvuetify import VuetifyTemplate
from solara.components.code_highlight_css import CodeHighlightCss

from domino_code_assist import util
from domino_code_assist.code import Code
from domino_code_assist.settings import settings
from domino_code_assist.snippets.snippets import Snippet


class TreeWidget(VuetifyTemplate):
    template_file = (__file__, "tree.vue")

    path = traitlets.List(default_value=[]).tag(sync=True)
    items = traitlets.List().tag(sync=True)
    scroll_pos = traitlets.Int(allow_none=True).tag(sync=True)
    refresh_count = traitlets.Int(0).tag(sync=True)


@reacton.component
def TreeBrowser(
    path: List,
    items: List,
    on_path: Callable[[List], None] = None,
    scroll_pos: int = None,
    on_scroll_pos: Callable[[int], None] = None,
    on_refresh_count: Callable[[int], None] = None,
):
    return TreeWidget.element(path=path, items=items, on_path=on_path, scroll_pos=scroll_pos, on_scroll_pos=on_scroll_pos, on_refresh_count=on_refresh_count)


@reacton.component
def SnippetsPanel(
    get_items: Callable[[List], List],
    get_content: Callable[[List], Snippet],
    path,
    on_path,
    edit_mode,
    on_edit_mode,
    writable_paths,
    on_edit_snippet,
    on_delete_snippet,
    on_add_snippet,
    on_snippet: Callable[[Optional[Snippet]], None] = lambda s: None,
    on_refresh_count: Callable[[int], None] = None,
):
    items, set_items = reacton.use_state(cast(List[Dict], []))
    content, set_content = reacton.use_state(cast(Optional[Snippet], None))

    scroll_pos_stack, set_scroll_pos_stack = reacton.use_state([0])  # pre add 1 item for the root
    scroll_pos, set_scroll_pos = reacton.use_state(0)

    delete_dialog, set_delete_dialog = reacton.use_state(False)
    add_dialog, set_add_dialog = reacton.use_state(False)
    add_name, set_add_name = reacton.use_state(None)
    add_path, set_add_path = reacton.use_state(None)

    def on_path_change():
        if not path[-1]["leaf"]:
            set_content(None)
            set_items(get_items(path))

            if len(path) >= len(scroll_pos_stack):
                set_scroll_pos_stack(scroll_pos_stack + [scroll_pos])
                set_scroll_pos(0)
            else:
                set_scroll_pos(scroll_pos_stack[len(path)])
                set_scroll_pos_stack(scroll_pos_stack[: len(path)])
        else:
            set_items(get_items(path[:-1]))
            set_content(get_content(path))

    reacton.use_memo(on_path_change, [path, get_items])

    def on_content():
        on_snippet(content)

    reacton.use_memo(on_content, [content])

    def is_writable(path: Path):
        def starts_with(a: Path, b: Path):
            try:
                b.relative_to(a)
                return True
            except ValueError:
                return False

        return any(starts_with(Path(p), path) for p in [wp["path"] for wp in writable_paths if wp["writable"]])

    def delete_snippet():
        on_delete_snippet(content)
        set_delete_dialog(False)

    with sol.Div(style_="height: 100%; display: flex; flex-direction: column;") as main:
        with sol.Div(style_="display: flex; flex-grow: 1; min-height: 0"):
            with sol.Div(style_="height: 100%; width: 50%; max-width: 50%;"):
                TreeBrowser(path, items, on_path, scroll_pos, set_scroll_pos, on_refresh_count)
            with sol.Div(style_="width: 50%; max-width: 50%; overflow: auto"):
                if content:
                    v.Html(tag="h3", class_="pt-2", children=[content.name.replace(".py", "")])
                    v.Html(tag="p", children=[content.description])
                    CodeHighlightCss()
                    Code(
                        code_chunks=[content.code],
                        on_event=lambda a: None,
                        error=None,
                    )
                else:
                    v.Html(tag="h3", class_="ma-4", children=["Please select a snippet to preview."])
        with sol.Div(style_="display: flex").meta(ref="button-bar"):
            v.Tooltip(
                bottom=True,
                v_slots=[
                    {
                        "name": "activator",
                        "variable": "tooltip",
                        "children": sol.Button(
                            v_on="tooltip.on",
                            icon_name="mdi-pencil" if edit_mode else "mdi-pencil-off",
                            color="primary" if edit_mode else "",
                            on_click=lambda: on_edit_mode(not edit_mode),
                            icon=True,
                        ).meta(ref="toggle-edit"),
                    }
                ],
                children=["Snippet editing enabled. Click to disable." if edit_mode else "Snippet editing disabled. Click to enable."],
            )
            if edit_mode:
                v.Tooltip(
                    bottom=True,
                    v_slots=[
                        {
                            "name": "activator",
                            "variable": "tooltip",
                            "children": sol.Button(
                                "add",
                                v_on="tooltip.on",
                                icon_name="mdi-plus",
                                color="primary",
                                on_click=lambda: set_add_dialog(True),
                                class_="mx-2",
                            ).meta(ref="btn-add"),
                        }
                    ],
                    children=["Add a snippet to the current folder."],
                )
                if content:
                    sol.Button(
                        "edit",
                        color="primary",
                        disabled=not is_writable(content.base_path),
                        on_click=lambda: on_edit_snippet(content),
                        icon_name="mdi-pencil",
                        class_="mx-2",
                    ).meta(ref="btn-edit")
                    sol.Button(
                        "delete",
                        color="primary",
                        disabled=not is_writable(content.base_path),
                        on_click=lambda: set_delete_dialog(True),
                        icon_name="mdi-trash-can",
                        class_="mx-2",
                    ).meta(ref="btn-delete")

        with v.Dialog(width="400px", v_model=delete_dialog, on_v_model=lambda v: set_delete_dialog(v)).meta(ref="dialog-delete"):
            with v.Card():
                v.CardTitle(children=["Are you sure you want to delete this snippet?"])
                with v.CardActions():
                    sol.Button("Cancel", outlined=True, color="primary", on_click=lambda: set_delete_dialog(False))
                    sol.Button("Delete", color="primary", on_click=delete_snippet, class_="mx-2")
        AddDialog(writable_paths, add_dialog, set_add_dialog, on_add_snippet, path)

    return main


@reacton.component
def AddDialog(writable_paths, open, on_open, on_add_snippet, path=[{}]):
    add_name, set_add_name = reacton.use_state("")
    add_path, set_add_path = reacton.use_state("")
    with v.Dialog(width="600px", v_model=open, on_v_model=lambda v: on_open(v)).meta(ref="add-dialog") as main:
        with v.Card():
            v.CardTitle(children=["Add snippet"])
            with v.CardText():
                sol.Div(children=["Folder: " + "/".join([p["id"] for p in path[1:] if not p["leaf"]])])
                v.TextField(label="Name", v_model=add_name, on_v_model=set_add_name, autofocus=True)
                sol.Div(children=["Repository"])
                with v.List():
                    with v.ListItemGroup(v_model=add_path, on_v_model=set_add_path, color="primary"):
                        for item in writable_paths:
                            label = item
                            icon = ""
                            git_path = settings.domino_repos_dir or settings.domino_imported_code_dir
                            if str(item["path"]).startswith(settings.domino_working_dir):
                                label = "current project"
                            elif git_path and str(item["path"]).startswith(git_path):
                                icon = "mdi-git"
                                label = str(item["path"])[len(git_path) + 1 :]
                            with v.ListItem(value=str(item["path"]), disabled=not item["writable"]).key(str(item)):
                                with v.ListItemIcon():
                                    if icon:
                                        v.Icon(children=[icon])
                                    else:
                                        v.Img(
                                            src=util.logo_grey,
                                            style_="height: 24px; max-width: 24px; border-radius: 12px",
                                        )
                                with v.ListItemContent():
                                    v.ListItemTitle(
                                        children=[
                                            str(label),
                                            sol.Tooltip(
                                                "Read only. The configured credentials don't have write access to this repository.",
                                                children=[sol.Text(" (read only)", style="pointer-events: all")],
                                            )
                                            if not item["writable"]
                                            else "",
                                        ]
                                    )
            with v.CardActions():
                sol.Button("Cancel", outlined=True, color="primary", on_click=lambda: on_open(False), icon_name="mdi-close")
                sol.Button(
                    "add",
                    color="primary",
                    on_click=lambda: on_add_snippet(add_path, path, add_name),
                    class_="mx-2",
                    disabled=not add_path or not add_name,
                    icon_name="mdi-plus",
                ).meta(ref="btn-add")
    return main



import dataclasses
from dataclasses import field
from pathlib import Path
from typing import Dict, Optional

import plotly
import solara
import solara.lab
from solara.lab import Ref

import domino_code_assist
from domino_code_assist.action import ActionDownloadDataSource

from .. import settings

options = [x for x in dir(plotly.data) if not x.startswith("_")] + [x for x in dir(domino_code_assist.data) if not x.startswith("_")]


@dataclasses.dataclass(frozen=True)
class LoadData:
    tab: int = 4
    data_source: ActionDownloadDataSource = ActionDownloadDataSource()
    dataset_dir: Path = field(default_factory=lambda: Path(settings.settings.domino_datasets_dir))
    dataset_file: Optional[str] = None
    project_dir: Path = field(default_factory=lambda: Path(settings.settings.domino_working_dir))
    project_file: Optional[str] = None
    upload_file: Optional[Dict] = None
    demo_df_name: str = "palmerpenguins"
    df_out: str = "df"
    reactive: bool = False

    def valid(self):
        if self.tab == 0:
            if self.data_source and self.data_source.type_ in ["S3Config", "GCSConfig"] and self.data_source.database:
                return True
            elif self.data_source.type_ == "BigQueryConfig":
                return bool(self.data_source and (self.data_source.query if self.data_source.use_query else self.data_source.schema and self.data_source.table))
            else:
                return bool(
                    self.data_source and (self.data_source.query if self.data_source.use_query else self.data_source.database and self.data_source.table)
                )
        if self.tab == 1:
            return bool(self.dataset_file)
        if self.tab == 2:
            return bool(self.project_file)
        if self.tab == 3:
            return bool(self.upload_file)
        if self.tab == 4:
            return True


class ReactiveDataSource(solara.lab.Reactive[ActionDownloadDataSource]):
    def set_type(self, type_):
        self.update(type_=type_, region="default" if type_ == "BigQueryConfig" else None, use_query=False)

    def set_name(self, name):
        self.update(name=name, database=None, schema=None, table=None, project=None)

    def set_database(self, value):
        self.update(database=value, schema=None, table=None)

    def set_schema(self, value):
        self.update(schema=value, table=None)

    def set_sample(self, value):
        if value != self.get().sample:
            self.update(sample=value)

    def set_project(self, value):
        self.update(project=value, schema=None, table=None)

    def set_region(self, value):
        self.update(region=value, schema=None, table=None)


class ReactiveLoadData(solara.lab.Reactive[LoadData]):

    data_source: ReactiveDataSource

    def __post__init__(self):
        self.data_source = ReactiveDataSource(Ref(self.fields.data_source))

    def on_dataset_dir_change(self, path: Path):
        self.update(dataset_dir=path, dataset_file=None)

    def on_project_dir_change(self, path: Path):
        self.update(project_dir=path, project_file=None)

    def on_dataset_file_change(self, path: Path):
        self.update(dataset_file=path)


load_data = ReactiveLoadData(LoadData())



import uuid
from dataclasses import asdict, dataclass, field
from typing import Dict, Optional

import solara.lab


class EqWrapper:
    def __init__(self, obj):
        self.obj = obj

    def get(self):
        return self.obj


@dataclass(frozen=True)
class Viz:
    name: str = "Untitled"
    plot_type: Optional[str] = None
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    col_args: Dict = field(default_factory=lambda: {})
    df_var_name: Optional[str] = None
    plot_var_name: Optional[str] = None
    crossfilter_enabled: bool = True


class ReactiveViz(solara.lab.Reactive[Viz]):
    pass


viz = ReactiveViz(Viz())


def to_json(obj):
    if isinstance(obj, Viz):
        return {"type_plot_state": asdict(obj)}


def from_json(dct):
    if "type_plot_state" in dct:
        res = Viz(**dct["type_plot_state"])
        return res



from typing import Any, Dict, cast

import nbformat
import reacton
import reacton.ipyvuetify as v
import solara as sol

from domino_code_assist import app, settings

cell_ast_list = []

nb: nbformat.NotebookNode = nbformat.read(".dca_deployed.ipynb", 4)
for cell_index, cell in enumerate(nb.cells):
    cell_index += 1  # used 1 based
    if cell.cell_type == "code":
        source = cell.source
        cell_path = f"{settings.settings.domino_notebook_deploy_filename} input cell {cell_index}"
        cell_ast_list.append(compile(source, cell_path, "exec"))


class ReferenceEqualityWrapper:
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        if not isinstance(other, ReferenceEqualityWrapper):
            return False
        return self.value is other.value


@reacton.component
def NbApp(handle_error: bool = True):
    replacements, set_replacements_ = reacton.use_state(cast(Dict[str, Any], {}), eq=lambda a, b: a is b)
    page, set_page = reacton.use_state(sol.Div())

    def set_replacement(key, value):
        set_replacements_(lambda x: {**x, key: value})

    app._replacements.provide((replacements, set_replacement))

    def exec_nb():
        user_ns: Dict[str, Any] = {}
        try:
            for ast in cell_ast_list:
                exec(ast, user_ns)
            set_page(user_ns["Page"]())
        except Exception:
            if handle_error:
                import traceback

                tb = traceback.format_exc()
                set_page(
                    sol.Div(
                        children=[
                            v.Html(tag="pre", children=[str(tb)]),
                            sol.Button(children=["Restart"], class_="ma-2", on_click=lambda: set_replacements_({})),
                        ]
                    )
                )
            else:
                raise

    reacton.use_memo(exec_nb, [ReferenceEqualityWrapper(replacements)])
    return page


@reacton.component
def Page():
    return NbApp()



import reacton
from react_ipywidgets.core import Element

from domino_code_assist import action

from .load_data.data_source import DataSourcePanel
from .transform.transform_action_ui import (
    FilterPanelView,
    GroupByPanelView,
    MergePanelView,
    SelectColumnsPanelView,
)

_ui_for_action = {
    action.ActionDownloadDataSource: DataSourcePanel,
    action.ActionFilter: FilterPanelView,
    action.ActionSelectColumns: SelectColumnsPanelView,
    action.ActionDropColumns: SelectColumnsPanelView,
    action.ActionGroupBy: GroupByPanelView,
    action.ActionMerge: MergePanelView,
}


@reacton.component
def EditAction(columns, dtypes, on_action, action_obj) -> Element:
    return _ui_for_action[action_obj.__class__](
        columns,
        dtypes,
        action_obj,
        on_action,
    )  # type: ignore



import contextlib
import dataclasses
import logging
import os
import shutil
import subprocess
import threading
import time
from datetime import datetime, timedelta, timezone
from enum import Enum
from pathlib import Path
from typing import Callable, Iterator, List, Optional, Tuple, cast

import humanize
import reacton.core
import solara
import solara.lab
import solara.util
from dateutil import parser  # type: ignore
from solara import Result
from solara.alias import reacton
from solara.alias import rv as v
from solara.alias import sol
from solara.components.applayout import should_use_embed

import domino_code_assist.assistant.notebook
from domino_code_assist.settings import settings

from ..assistant import mixpanel
from ..domino_api import AppInfo, get_domino_api
from .app_checker import AppChecker
from .button_clipboard import ButtonClipboard
from .filesystem_watcher import FileSystemWatcher

run_sh_content_template = """#!/usr/bin/env bash
SOLARA_APP=domino_code_assist.app_entrypoint SOLARA_TELEMETRY_MIXPANEL_ENABLE=False SOLARA_THEME_VARIANT={theme_variant}\\
    SOLARA_THEME_LOADER={theme_loader} uvicorn solara.server.starlette:app\\
    --loop asyncio\\
    --host=0.0.0.0 --port=8888"""


logger = logging.getLogger("dca.deploy")

deploy_sync = solara.util.load_file_as_data_url(Path(__file__).parent / "deploy-sync.png", "image/png")

# these are changed from test tests
timeout_notebook_save: float = 60 * 3
timeout_filesystem_sync: float = 60 * 1
timeout_app_stop: int = 60
timeout_app_start: int = 60 * 5
delay_app_state: float = 10.0
delay_app_publish: float = 3.0
app_poll_delay: float = 1
server_log_update_deplay: float = 10.0
fresh_deploy_delta = timedelta(minutes=1)


class AppStatus(str, Enum):
    NON_EXISTENT = "Non existent"  # not a return value by the domino api, but used interally
    PENDING = "Pending"
    PREPARING = "Preparing"
    RUNNING = "Running"
    STOPPED = "Stopped"
    QUEUED = "Queued"
    FAILED = "Failed"
    SUCCEEDED = "Succeeded"
    ERROR = "Error"


APP_END_STATES = (AppStatus.NON_EXISTENT, AppStatus.STOPPED, AppStatus.FAILED, AppStatus.RUNNING, AppStatus.SUCCEEDED, AppStatus.ERROR)


class AppOperation(str, Enum):
    START = "Start"
    STOP = "Stop"


notebook_saved_error = solara.lab.Reactive(False)
notebook_save_request = solara.lab.Reactive(0)
deploy_step = solara.lab.Reactive("")


@dataclasses.dataclass
class LogMessage:
    message: str
    timestamp: datetime = dataclasses.field(default_factory=lambda: datetime.now())


deploy_log = solara.lab.Reactive(cast(Tuple[LogMessage, ...], ()))


@contextlib.contextmanager
def step(info: str):
    deploy_step.value = info
    try:
        log("## " + info)
        yield
    except Exception as e:
        # deploy_error.value = e
        log("faild step: " + info + " with error: " + str(e))
        raise


def log(message: str):
    msg = LogMessage(message)
    # useful when debugging with the tests to see this in stdout
    # print(msg.timestamp, msg.message, flush=True)
    deploy_log.value += (msg,)


def validate_app():
    proc = subprocess.run(
        ["python", "-m", "domino_code_assist.app_entrypoint"], cwd=settings.domino_working_dir, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    return proc


def track_view_app():
    mixpanel.api.track_with_defaults(
        "interaction",
        {
            "section": "deploy",
            "type": "view app",
        },
    )


def deploy(notebook_saved: threading.Event, filesytem_synced: threading.Event, notebook_browser_path: str, theme_variant="light"):
    mixpanel.api.track_with_defaults(
        "interaction",
        {
            "section": "deploy",
            "type": "deploy start",
        },
    )
    with step("Generating app startup script"):
        log("Generating app startup script content using theme variant: " + theme_variant)
        run_sh_content = run_sh_content_template.format(
            theme_variant=theme_variant,
            theme_loader="plain",
        ).encode("utf8")
        log(f"Write app.sh to working dir: {settings.domino_working_dir}")
        root = Path(settings.domino_working_dir)
        path = root / "app.sh"

        with open(path, "wb") as f:
            f.write(run_sh_content)
        log("Generated app startup script")

    with step("Saving notebook"):
        log("Requesting notebook save")
        notebook_save_request.value += 1

        if not notebook_saved.wait(timeout=timeout_notebook_save):
            log(f"Notebook save timeout: waited {timeout_notebook_save} seconds, but got no response")
            raise TimeoutError("Could not save notebook")
        if notebook_saved_error.value:
            log("Notebook saved, but return an error")
            raise Exception("Could not save notebook")
        root = Path(settings.domino_working_dir)
        notebook_local_path = Path(os.getcwd()) / Path(notebook_browser_path).name
        log(f"Saved notebook to {notebook_local_path}")
        shutil.copy2(notebook_local_path, root / settings.domino_notebook_deploy_filename)
        log(f"Copied notebook for deployment to {root / settings.domino_notebook_deploy_filename}")

    with step("Validating app"):
        log(f"Running python -m domino_code_assist.app_entrypoint with working directory: {settings.domino_working_dir}")
        proc = validate_app()
        if proc.returncode != 0:
            raise Exception(proc.stderr.decode("utf-8").split("\n")[-2])
        log("Notebook ran without errors")

    with step("Syncing files"):
        # what if not file changes?
        log("Asking Domino API endpoint to sync Project Files")
        get_domino_api().sync("Auto sync from Domino Code Assist for deployment")

        if not filesytem_synced.wait(timeout=timeout_filesystem_sync):
            log(f"Did not get a confirmation that files were synced within {timeout_filesystem_sync} seconds")
            raise TimeoutError("Did not synchronize files")
        log("Files were synced")

    with step("Deploying new app... (this may take a while)"):
        log("Asking Domino API endpoint for app status")
        # what if we started a new project?
        status = get_domino_api().get_app_status() or AppStatus.NON_EXISTENT
        if status not in APP_END_STATES:
            log(f"App status {status} is not in {APP_END_STATES}")
            raise Exception(f"Unexpected initial state: {status} for app")
        log(f"Got initial app state: {status}")

        if status == AppStatus.RUNNING:
            log("Asking Domino API endpoint to stop app")
            get_domino_api().app_unpublish()

            for i in range(timeout_app_stop):
                status = get_domino_api().get_app_status() or AppStatus.NON_EXISTENT
                if status == AppStatus.STOPPED:
                    log("App stopped")
                    break
                log(f"App is in state {status}, waiting for it to stop")
                time.sleep(app_poll_delay)
            else:
                log(f"App did not stop within {timeout_app_stop} seconds")
                raise TimeoutError("Could not stop app")

        log("Asking Domino API endpoint to start app")
        # If we publish directly, maybe the app will not pick up the latest file changes
        # To be safe, we wait a bit
        time.sleep(delay_app_publish)
        get_domino_api().app_publish()
        for i in range(timeout_app_start):
            status = get_domino_api().get_app_status() or AppStatus.NON_EXISTENT
            if status == AppStatus.RUNNING:
                log("App started")
                mixpanel.api.track_with_defaults(
                    "interaction",
                    {
                        "section": "deploy",
                        "type": "deploy success",
                    },
                )
                return True
            elif status in APP_END_STATES:
                log(f"App is in state {status}, but expected it to be {AppStatus.RUNNING}")
                # raise Exception(f"Could not start app (status: {status})")
                return False
            log(f"App is in state {status}, waiting for it to start")
            time.sleep(app_poll_delay)
        else:
            log(f"App did not start within {timeout_app_start} seconds")
            raise TimeoutError(f"Could not start app (status: {status})")


@solara.component
def Advanced():
    deploy_log.use()
    # server_logs, set_server_logs = solara.use_state([""])

    def get_logs() -> Iterator[Optional[List[str]]]:
        while True:
            if get_domino_api().get_app_id() is not None:
                logs = get_domino_api().get_app_log()
                lines = [k["log"] for k in logs]
                yield lines
            else:
                yield None
            time.sleep(server_log_update_deplay)

    server_log_result: Result[List[str]] = solara.use_thread(get_logs)

    lines = [f"{k.timestamp}: {k.message}" for k in deploy_log.value]
    text = "\n".join(lines)

    with v.Tabs() as main:
        v.Tab(children=["DCA deploy logs"])
        v.Tab(children=["Domino server app logs"])
        with v.TabItem():
            solara.Preformatted(text).meta(ref="log")
        with v.TabItem():
            if server_log_result.error:
                solara.Error(repr(server_log_result.error))
            else:
                server_log = "\n".join(server_log_result.value or ["No app exists yet"])

                solara.Style(
                    """
                .dca-server-logs {
                    overflow: auto;
                }
                """
                )
                solara.Text("Logs update every 10 seconds")
                solara.Preformatted(server_log, attributes={"class": "dca-server-logs"}).meta(ref="server_logs")
                url = f"/launchpad-publisher/{settings.domino_project_owner}/{settings.domino_project_name}/publishApp"
                solara.HTML(unsafe_innerHTML=f'Watch the logs at <a href="{url}" target="_blank">The Domino interface</a>.')
    return main


def use_current_environment() -> Result[str]:
    def get_current_environment():
        current_id = get_domino_api().get_current_environment_id()
        environments = get_domino_api().get_useable_environments()["environments"]
        names = [k["name"] for k in environments if k["id"] == current_id]
        if len(names) == 1:
            return names[0]
        else:
            return "UNKNOWN"

    return solara.use_thread(get_current_environment)


def use_default_environment() -> Result[str]:
    def get_default_environment():
        env_info = get_domino_api().get_useable_environments()
        environments = env_info["environments"]
        default_id = env_info["currentlySelectedEnvironment"]["id"]
        names = [k["name"] for k in environments if k["id"] == default_id]
        if len(names) == 1:
            return names[0]
        else:
            return "UNKNOWN"

    return solara.use_thread(get_default_environment)


@reacton.component
def Deployer(app: Callable[[], reacton.core.Element]):
    successfull_once, set_successfull_once = solara.use_state(False)
    show_app_deployed_dialog, set_show_app_deployed_dialog = solara.use_state(False)
    open_advanced, set_open_advanced = solara.use_state(False)
    online, set_online = reacton.use_state(False)
    open_preview, set_open_preview = reacton.use_state(False)
    deploy_request, set_deploy_request = reacton.use_state(False)
    filesystem_watcher_connected, set_filesystem_watcher_connected = reacton.use_state(False)
    last_deploy, set_last_deploy = solara.use_state(cast(Optional[datetime], None))
    filesystem_sync_failed, set_filesystem_sync_failed = reacton.use_state(False)

    current_env: Result[str] = use_current_environment()
    default_env: Result[str] = use_default_environment()

    def update_last_deploy():
        if deploy_request and not show_app_deployed_dialog:
            # we closed the dialog
            set_last_deploy(datetime.now())

    solara.use_memo(update_last_deploy, [deploy_request, show_app_deployed_dialog])

    notebook_save_request.use()
    domino_code_assist.assistant.notebook.notebook_browser_path.use()
    deploy_step.use()

    # The reactive variables need to be reset when the component gets re-created
    # since the lifetime is tied to the Python runtime.
    def reset():
        deploy_step.value = ""
        notebook_saved_error.value = False
        deploy_log.value = ()

    solara.use_memo(reset, [])

    # We constantly poll the app status, since it may sometimes fail later on (after deployment is successful)
    # or a different process/UI may change the app state.

    def app_status_info_run(cancel: threading.Event) -> Iterator[Tuple[AppStatus, AppInfo]]:
        while True:
            try:
                status = get_domino_api().get_app_status() or AppStatus.NON_EXISTENT
                info = get_domino_api().get_app_info()
                yield status, info[0] if len(info) else None
            except Exception:
                logger.exception("Error getting app status or info")
            time.sleep(delay_app_state)

    app_state: Result[Tuple[AppStatus, AppInfo]] = sol.use_thread(app_status_info_run, dependencies=[])
    # split app_state into app_status and app_info
    app_status: Optional[AppStatus] = None
    app_info: Optional[AppInfo] = None
    if app_state.state == solara.ResultState.RUNNING and app_state.value is not None:
        assert app_state is not None
        app_status, app_info = app_state.value

    # We run the deploy in a thread, and communicate via the reactive variables,
    # the return value (True if successful, False otherwise) and two events (notebook_saved and filesystem_synced)
    def run_deploy():
        if deploy_request:
            return deploy(
                notebook_saved=notebook_saved,
                filesytem_synced=filesytem_synced,
                notebook_browser_path=domino_code_assist.assistant.notebook.notebook_browser_path.value,
            )

    # these all have the same dependencies, such that if we 'redo' a deploy, we reset all of it
    notebook_saved = solara.use_memo(threading.Event, dependencies=[deploy_request])
    filesytem_synced = solara.use_memo(threading.Event, dependencies=[deploy_request])
    deployed: Result[bool] = sol.use_thread(run_deploy, dependencies=[deploy_request])

    # We respond to the notebook save request using a side effect
    # which will then trigger an notebook_saved event.
    initial_notebook_saved_value = solara.use_memo(notebook_save_request.get, dependencies=[])

    def save_notebook():
        from domino_code_assist.assistant import notebook

        assert notebook.save_notebook

        def on_save(is_saved):
            notebook_saved_error.value = not is_saved
            notebook_saved.set()

        # we don't want to save the notebook on the first run
        if initial_notebook_saved_value != notebook_save_request.value:
            notebook.save_notebook(on_save)

    solara.use_effect(save_notebook, dependencies=[notebook_save_request.value])

    app_url = f"/modelproducts/{app_info['id']}" if app_info else ""
    disable_view = (deploy_request and not online) or (not deploy_request and app_status != AppStatus.RUNNING)
    should_use_embed.provide(False)
    with v.Sheet() as main:
        with v.Dialog(v_model=open_preview, on_v_model=set_open_preview, fullscreen=True, hide_overlay=True, persistent=True, no_click_animation=True):
            with v.Sheet(class_="overflow-y-auto overflow-x-auto"):
                with solara.AppLayout():
                    with solara.AppBar():
                        solara.Button(icon_name="mdi-close", on_click=lambda: set_open_preview(False), icon=True)
                    if open_preview:
                        app()
        with sol.Div(class_="d-flex justify-center"):

            def deploy_or_retry():
                # WARNING: in the future, if reacton batches updates, we may have to find another way
                # as toggling a boolean might not trigger a re-render
                if deployed.error:
                    set_deploy_request(False)
                    set_deploy_request(True)
                else:
                    if deploy_request:
                        set_deploy_request(False)
                    set_deploy_request(True)
                # make sure we start with the default valyue again
                set_online(False)

            loading = False
            if deploy_request:
                if deployed.state == solara.ResultState.RUNNING:
                    loading = True
                elif deployed.state == solara.ResultState.FINISHED and app_status == AppStatus.RUNNING and not deployed.error:
                    loading = not online
            else:
                loading = (not filesystem_watcher_connected) or app_status not in APP_END_STATES
            solara.Button(
                "Re-Deploy app" if successfull_once else "Deploy app",
                on_click=deploy_or_retry,
                color="primary",
                disabled=loading,
                class_="ma-1",
                icon_name="mdi-rocket",
            ).meta(ref="deploy")
            sol.Button(
                "View app",
                class_="ma-1",
                color="primary",
                icon_name="mdi-application",
                disabled=disable_view,
                href=app_url,
                target="_blank",
                on_click=track_view_app,
                outlined=True,
            ).meta(ref="view_app")
            ButtonClipboard(pathname=app_url, disabled=disable_view)
            solara.Button(
                "Preview app",
                icon_name="mdi-magnify",
                color="primary",
                on_click=lambda: set_open_preview(True),
                outlined=True,
                class_="ma-1",
            )
            v.Spacer()
            sol.Button(
                "Advanced",
                class_="ma-1",
                icon_name="mdi-cogs",
                on_click=lambda: set_open_advanced(True),
                text=True,
            ).meta(ref="advanced")
        with v.Dialog(v_model=open_advanced, on_v_model=set_open_advanced, width="800px"):
            with v.Card():
                with v.CardText(class_="pt-2"):
                    if open_advanced:
                        Advanced()
                with v.CardActions():
                    sol.Button("Close", icon_name="mdi-close", on_click=lambda: set_open_advanced(False)).meta(ref="close_advanced")

        message_type = "info"
        if app_state.state == solara.ResultState.ERROR:
            info_message = "Error getting app state"
            message_type = "error"
        elif app_status is None:
            info_message = "Getting app state..."
        else:
            info_message = f"App state: {app_status}"
            if deployed.state == solara.ResultState.ERROR:
                info_message = f"Error deploying app: {deployed.error}"
                message_type = "error"
            else:
                if app_status == AppStatus.RUNNING and deployed.state == solara.ResultState.FINISHED and filesystem_watcher_connected:
                    if app_info:
                        try:
                            last_deploy_msg = humanize.naturaltime(datetime.now(timezone.utc) - parser.parse(app_info["created"]))
                        except Exception as e:
                            last_deploy_msg = f"error: {type(e).__name__} - {str(e)}"
                        info_message = (
                            f"The app was deployed {last_deploy_msg}. "
                            'If you\'ve made changes since the last deployment, click the "DEPLOY APP" button to redeploy.'
                        )
                    else:
                        info_message = "No app deployed yet"
                elif deployed.state == solara.ResultState.RUNNING:
                    if deploy_request:
                        # we get the info from the deploy function via deploy_step
                        info_message = deploy_step.value
                        if not info_message:
                            info_message = "Deploying app..."
        if not filesystem_watcher_connected:
            info_message = "Waiting for file system watcher..."

        if deployed.value and deployed.state == solara.ResultState.FINISHED and app_status == AppStatus.RUNNING:
            if online:
                fresh_deploy = False
                if last_deploy is not None:
                    fresh_deploy = (datetime.now() - last_deploy) <= fresh_deploy_delta
                if fresh_deploy:
                    info_message = "App deployed"
                    message_type = "success"
                # else we keep the "The app was deployed ... ago message"
            else:
                info_message = "Waiting for app to get online"
        if deployed.value is False and deployed.state == solara.ResultState.FINISHED:
            info_message = "App not deployed, click the advanced button for more details."
            message_type = "error"
        v.ProgressLinear(indeterminate=True, style_="visibility: " + ("hidden" if not loading else "visible"), class_="ma-1")
        v.Alert(type=message_type, children=[info_message], text=True, outlined=True, dense=False, icon=True, class_="ma-1").meta(ref="info_message")

        # only if the deployment if not successful, we show a warning when the environment is different
        if current_env.state == solara.ResultState.FINISHED and default_env.state == solara.ResultState.FINISHED and current_env.value != default_env.value:
            with solara.Warning(
                f"The current workspace is using the {current_env.value!r} environment, but the app is deployed with the {default_env.value!r} environment. "
                "This may have caused your app to fail",
                classes=["ma-1"],
            ).meta(ref="environment_message").key("environment_message"):
                settings_urls = f"/u/{settings.domino_project_owner}/{settings.domino_project_name}/settings#execution"
                solara.Markdown(
                    f'<a href="{settings_urls}" target="_blank">Changing your default environment to {current_env.value!r}'
                    "in your settings may solve this problem.</a>"
                )

        if deployed.value is not False:
            # if deployed, we do not need the filesystem watcher anymore
            # the frontend knows the hosts, we don't
            server_path = f"/u/{settings.domino_project_owner}/{settings.domino_project_name}/interactiveSession/ws/{settings.domino_run_id}"

            def on_message(msg):
                if msg is None:
                    return
                type = msg.get("type")
                if type == "RunResourcesStatusChangedEvent":
                    status = msg.get("status", {})
                    repositories = status.get("repositories", [])
                    for repository in repositories:
                        name = repository.get("name", "")
                        if name == "Domino Files":
                            status = repository.get("status", "")
                            clean = status.get("Clean")
                            if clean == {}:
                                if not filesytem_synced.is_set():
                                    log("Filesystem is clean, no need to sync")
                                filesytem_synced.set()
                                return
                elif type == "RunFullFileSynchronizationEvent":
                    key = msg.get("key", "")
                    if key == "full.sync.succeeded":
                        if not filesytem_synced.is_set():
                            log("Filesystem is synced successfully")
                        filesytem_synced.set()
                        set_filesystem_sync_failed(False)
                        return
                elif type == "FailedFileSynchronizationEvent":
                    key = msg.get("key", "")
                    if key == "full.sync.failed":
                        log("Filesystem sync failed")
                        set_filesystem_sync_failed(True)
                        return

            FileSystemWatcher(
                server_path=server_path, on_connected=set_filesystem_watcher_connected, on_message=on_message, dev=bool(settings.domino_code_assist_dev)
            ).meta(ref="filesystem_watcher").key("filesystem_watcher")

        def on_online(value):
            log("App is online" if value else "App is offline")
            set_online(value)
            # should_show_app_deployed_dialog = deploy_request and online and app_status == AppStatus.RUNNING and deployed.state == solara.ResultState.FINISHED
            # we only show it once
            if value and deploy_request and not show_app_deployed_dialog and deployed.state == solara.ResultState.FINISHED:
                set_show_app_deployed_dialog(True)
                set_successfull_once(True)

            # solara.use_memo(update_popup, [deploy_request, online, app_status, deployed.state])

        if bool(deployed.value) or bool(app_info):
            AppChecker(
                url=app_info["openUrl"] if app_info else "",
                online=online,
                running=app_status == AppStatus.RUNNING,
                on_online=on_online,
                dev=bool(settings.domino_code_assist_dev),
            ).key("app_checker")

        # popup when deployment is succesfull
        with v.Dialog(v_model=show_app_deployed_dialog, on_v_model=lambda v: set_show_app_deployed_dialog(v), width="400px").meta(ref="popup").key("popup"):
            with v.Card():
                with v.CardText(class_="pt-2"):
                    sol.Info(sol.HTML(unsafe_innerHTML="<h2>Your app is deployed! 🎉<h2>"), icon=False, classes=["mt-4", "mb-0"])

                with v.CardActions(class_="mx-3"):
                    with sol.Div(class_="d-flex justify-center"):
                        sol.Button(
                            "View app",
                            color="primary",
                            class_="ma-1",
                            icon_name="mdi-application",
                            href=app_url,
                            target="_blank",
                            on_click=track_view_app,
                        )
                        ButtonClipboard(pathname=app_url)
                    sol.Button(
                        "Close",
                        icon_name="mdi-close",
                        on_click=lambda: set_show_app_deployed_dialog(False),
                        color="primary",
                        outlined=True,
                    )
        if filesystem_sync_failed:
            with solara.Error(
                "File system sync failed. Please try to sync again manually using the Domino interface, as shown in the image below.", classes=["ma-1"]
            ).meta(ref="sync_error").key("sync_error"):
                v.Img(src=deploy_sync, style_="height: 455px", contain=True, class_="ma-4").meta(ref="sync_image")

    return main



import logging.config

import reacton

from .nb_app import NbApp

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "()": "uvicorn.logging.DefaultFormatter",
            "fmt": "%(levelprefix)s %(name)s %(message)s",
            "use_colors": None,
        },
    },
    "handlers": {
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
    },
    "loggers": {
        "": {"handlers": ["default"], "level": "WARNING", "propagate": False},
    },
}

logging.config.dictConfig(LOGGING_CONFIG)


@reacton.component
def Page(handle_error=True):
    return NbApp(handle_error)


if __name__ == "__main__":
    import reacton

    reacton.render(Page(handle_error=False), handle_error=False)



import ipyvuetify as v
import reacton
import traitlets


class ButtonClipboardWidget(v.VuetifyTemplate):
    template_file = (__file__, "button_clipboard.vue")

    pathname = traitlets.Unicode("").tag(sync=True)
    show_clipboard_message = traitlets.Bool(False).tag(sync=True)
    disabled = traitlets.Bool(False).tag(sync=True)


@reacton.component
def ButtonClipboard(pathname: str, disabled: bool = False):
    return ButtonClipboardWidget.element(pathname=pathname, disabled=disabled)



from typing import Callable

import ipyvuetify as iv
import reacton
import traitlets


class AppCheckerWidget(iv.VuetifyTemplate):
    template_file = (__file__, "app_checker.vue")

    url = traitlets.Unicode("").tag(sync=True)
    online = traitlets.Bool(False).tag(sync=True)
    running = traitlets.Bool(False).tag(sync=True)
    dev = traitlets.Bool(False).tag(sync=True)


@reacton.component
def AppChecker(url: str, online: bool, running: bool, on_online: Callable[[bool], None], dev: bool):
    return AppCheckerWidget.element(url=url, online=online, running=running, on_online=on_online, dev=dev)



import site
import sys
from pathlib import Path
from typing import Optional

import pydantic

prefix = Path(sys.prefix)
PROJECT_ROOT = Path(__file__).resolve().parent.parent
in_git = (PROJECT_ROOT / ".git").exists()

if in_git:
    # for development we use the prefix directory instead of the real prefix
    prefix = PROJECT_ROOT / "prefix"

if __file__.startswith(site.getuserbase()):
    # when installed in user mode, we use the user directory
    prefix = Path(site.getuserbase())

domino_snippet_builtin_dir_default = prefix / "share" / "domino-code-assist" / "snippets"


class Settings(pydantic.BaseSettings):
    domino_code_assist_dev: Optional[bool]
    domino_code_assist_exception_guard: bool = True
    snowflake_user: Optional[str]
    snowflake_password: Optional[str]
    snowflake_account: Optional[str]

    redshift_host: Optional[str]
    redshift_user: Optional[str]
    redshift_password: Optional[str]

    domino_api_class: Optional[str]

    domino_user_api_key: Optional[str]
    domino_project_id: Optional[str]
    domino_project_owner: Optional[str]
    domino_project_name: Optional[str]
    domino_api_host: Optional[str]
    domino_alternative_api_host: Optional[str]
    domino_run_id: Optional[str]
    domino_hardware_tier_id: Optional[str]

    domino_working_dir: str
    domino_datasets_dir: str = "/domino/datasets/"

    domino_snippet_builtin_dir: Path = domino_snippet_builtin_dir_default

    path_translated: Optional[str]
    domino_repos_dir: Optional[str]
    domino_imported_code_dir: Optional[str]
    domino_notebook_deploy_filename = ".dca_deployed.ipynb"

    domino_is_git_based: Optional[bool]

    class Config:
        case_sensitive = False
        env_file = ".env"


settings = Settings()



from typing import Optional

import reacton
import reacton.ipyvuetify as v
from solara.lab import Ref

from .state import load_data

_items = {
    "Quick-start": ["carshare", "election", "iris", "medals_long", "medals_wide", "palmerpenguins", "tips", "wind"],
    "Economics": ["gapminder", "stocks", "stocks_long"],
    "Pharma": ["experiment", "small_molecule_drugbank"],
}


@reacton.component
def QuickStart():
    load_data.use()
    demo_df_name = Ref(load_data.fields.demo_df_name)

    def find_selected_index(df_name: str) -> Optional[int]:
        for i, (category, items) in enumerate(_items.items()):
            if df_name in items:
                return i
        return None

    selected_index = find_selected_index(demo_df_name.value)

    expanded, set_expanded = reacton.use_state(selected_index)

    with v.ExpansionPanels(v_model=expanded, on_v_model=set_expanded, accordeon=True, style_="width: 800px") as main:
        for i, category in enumerate(_items.keys()):
            with v.ExpansionPanel():
                v.ExpansionPanelHeader(
                    children=[
                        v.Html(tag="strong", children=[f"{category} ({demo_df_name.value})"]) if selected_index == i and expanded != i else category,
                    ]
                )
                with v.ExpansionPanelContent():
                    with v.List():
                        selected = demo_df_name.get() if demo_df_name.get() in _items[category] else None
                        with v.ListItemGroup(
                            color="primary", v_model=selected, on_v_model=lambda v: demo_df_name.set(v) if v else None, mandatory=expanded == i
                        ):
                            for item in _items[category]:
                                with v.ListItem(value=item).key(item + "|list"):
                                    with v.ListItemAvatar():
                                        v.Icon(children=["mdi-check-box-outline" if demo_df_name.get() == item else "mdi-checkbox-blank-outline"])
                                    with v.ListItemContent():
                                        v.ListItemTitle(children=[item])
    return main



import os
import re
from keyword import iskeyword
from typing import Any, Callable, Dict, List, Optional

import nbformat as nbf
import reacton
import reacton.ipyvuetify as v
import solara.util as sutil
from IPython import get_ipython

from .settings import settings


def get_nb_locals():
    def get_frame():
        try:
            assert False
        except Exception as e:
            return e.__traceback__.tb_frame if e.__traceback__ is not None else None

    def find_nb_locals(current):
        while current is not None:
            if current.f_locals.get("__doc__") == "Automatically created module for IPython interactive environment":
                return current.f_locals

            current = current.f_back

    return find_nb_locals(get_frame())


nb_locals = get_nb_locals()


def get_vars(filter: Callable[[Any], bool]):
    return [k for k, v in nb_locals.items() if not re.search(r"^_.*$", k) and k != "Page" and filter(v)]


def generate_var_name(prefix):
    count = 0
    while True:
        count += 1
        name = f"{prefix}_{count}"
        if [x for x in get_vars(lambda v: True) if x == name]:
            continue
        return name


logo = sutil.load_file_as_data_url(os.path.join(os.path.dirname(__file__), "DominoLogoWhite.svg"), "image/svg+xml")

logo_grey = sutil.load_file_as_data_url(os.path.join(os.path.dirname(__file__), "DominoLogoGrey.svg"), "image/svg+xml")


@reacton.component
def DominoHeader(title, show_logo=True):
    main = v.CardTitle(
        class_="text-h5",
        style_="background-color: #2d71c7; color: white",
        children=[
            title,
            *(
                [
                    v.Spacer(),
                    v.Img(
                        src=logo,
                        style_="height: 32px; max-width: 32px; border-radius: 16px",
                    ),
                ]
                if show_logo
                else []
            ),
        ],
    )
    return main


def eq_by_ref(a, b):
    return a is b


def copy_upload(file_obj, name):
    import errno
    import shutil
    from datetime import datetime

    now = datetime.now()
    dt_string = now.strftime("%Y%m%d%H%M%S")
    data_dir = os.path.join(settings.domino_working_dir, "domino_code_assist_data")
    try:
        os.makedirs(data_dir)
    except OSError as e:
        if not e.errno == errno.EEXIST:
            raise
    dest_name = os.path.join(data_dir, dt_string + "_" + name)
    destination = open(dest_name, "wb")
    shutil.copyfileobj(file_obj, destination)

    return dest_name


def python_safe_name(name, used=None):
    if used is None:
        used = []
    first, rest = name[0], name[1:]
    name = re.sub("[^a-zA-Z_]", "_", first) + re.sub("[^a-zA-Z_0-9]", "_", rest)
    if name in used:
        nr = 1
        while name + ("_%d" % nr) in used:
            nr += 1
        name = name + ("_%d" % nr)
    return name


def in_dev_mode():
    return settings.domino_code_assist_dev


def remove_keys(d: Dict, keys: List[str]):
    return {k: v for k, v in d.items() if k not in keys}


def is_valid_variable_name(name):
    return name.isidentifier() and not iskeyword(name)


def store_notebook(file_name: str):
    ipython = get_ipython()
    nb = nbf.v4.new_notebook()
    nb["cells"] = [nbf.v4.new_code_cell(code) for code in ipython.history_manager.input_hist_raw]
    nbf.write(nb, file_name)


def mlflow_log_notebook(run_name: str):
    file_name = f"{run_name}_run_log.ipynb"
    store_notebook(file_name)

    try:
        from mlflow import log_artifact

        log_artifact(file_name)
    finally:
        os.remove(file_name)


def to_float(s: Optional[str]):
    if s is None:
        return None
    try:
        return float(s)
    except ValueError:
        return None


def to_int(s: Optional[str]):
    if s is None:
        return None
    try:
        return int(s)
    except ValueError:
        return None



from dataclasses import replace
from typing import Optional, cast

import reacton
import reacton.ipyvuetify as v
import solara as sol

import domino_code_assist.actions_store as actions_store
import domino_code_assist.util
from domino_code_assist import action, data, util
from domino_code_assist.assistant import drawer, mixpanel
from domino_code_assist.df_select import DfSelect
from domino_code_assist.hooks import use_reducer_addon
from domino_code_assist.serialize import dumps

from .transform_panel import TransformPanel


@reacton.component
def TransformDrawer(action_state: Optional[actions_store.ActionsState], set_code, dfs, open, edit, on_close, on_load_own_data, overwrite_warning=None):
    state, set_state = sol.use_state_or_update(action_state if open and edit and action_state else actions_store.ActionsState(preview=True))
    dispatch = use_reducer_addon(actions_store.actions_reducer, set_state)
    df_var_name, set_df_var_name = reacton.use_state(
        cast(Optional[str], (state.actions and state.actions[0] and cast(action.ActionUseNbLocals, state.actions[0]).var_name) or None)
    )

    actions_store.use_execute_df(state, set_state, domino_code_assist.util.nb_locals)

    quick_start_created, set_quick_start_created = reacton.use_state(False)

    def set_transformations():
        if df_var_name == "df_quick_start" and quick_start_created:
            set_quick_start_created(False)
            new_action = action.ActionDemo(df_var_out="df_quick_start", df_name="palmerpenguins")
            set_code({"code": new_action.render_code("df") + f"\n{new_action.df_var_out}", "meta": dumps(new_action), "modifier": "insert-above"})
            set_quick_start_created(False)

        code_str = action.render_code(action.render_code_chunks(state.actions))
        set_code({"code": code_str, "meta": dumps(state)})
        mixpanel.api.track_with_defaults("inserted code", {"section": "transformations"})
        close()

    def close():
        on_close()
        if quick_start_created:
            set_quick_start_created(False)
            util.nb_locals.pop("df_quick_start", None)
        set_df_var_name(None)
        set_state(actions_store.ActionsState(preview=True))

    def on_df_var_change():
        if df_var_name:
            if df_var_name == "Load more data":
                close()
                on_load_own_data()
                return
            if df_var_name == "df_quick_start" and df_var_name not in dfs:
                # pre populate notebook scope
                set_quick_start_created(True)
                util.nb_locals["df_quick_start"] = data.palmerpenguins()

            set_state(
                lambda state: replace(
                    state,
                    actions=[action.ActionUseNbLocals(var_name=df_var_name, df_var_out=df_var_name or "mypy")]
                    + (state.actions[1:] if state.actions else []),  # type: ignore
                )
            )
            mixpanel.api.track_with_defaults(
                "interaction",
                {
                    "section": "transformations",
                    "type": "selected dataframe",
                },
            )

    reacton.use_memo(on_df_var_change, [df_var_name])

    extended_dfs = dfs
    if len(dfs) == 0:
        extended_dfs = ["df_quick_start"] + extended_dfs
    if not edit:
        extended_dfs = extended_dfs + [{"divider": True}, "Load more data"]

    def pre_select_df():
        if open and (df_var_name is None or df_var_name not in dfs):
            set_df_var_name(dfs[0] if dfs else "df_quick_start")

    reacton.use_memo(pre_select_df, [open, dfs])

    with drawer.RightDrawer(
        open=open,
        on_open=lambda v: close() if not v else None,
        title="Transformations",
        edit=bool(edit),
        on_apply=set_transformations,
        warning_widget=overwrite_warning,
    ) as main:
        should_select_var = not df_var_name and not (state.actions and edit)
        with sol.Div(class_="dca-transform-header", style_="margin-left: -24px; margin-right: -24px; margin-top: -4px; padding: 8px 80px;"):
            DfSelect(value=df_var_name, label="DataFrame", on_value=set_df_var_name, items=extended_dfs or [])
        with sol.Div(style_="margin-left: -24px; margin-right: -24px"):
            v.Divider()
        if not should_select_var:
            TransformPanel(state, dispatch, on_reset=lambda: None, on_save=lambda: None, breakpoints=False, assistant_mode=True).meta(ref="TransformPanel")
    return main



from typing import Any, Callable

import ipyvuetify as iv
import reacton
import traitlets


class FileSystemWatcherWidget(iv.VuetifyTemplate):
    template_file = (__file__, "filesystem_watcher.vue")

    server_path = traitlets.Unicode("").tag(sync=True)
    dev = traitlets.Bool(False).tag(sync=True)
    on_connected = traitlets.Callable()
    on_message = traitlets.Callable()

    def vue_on_message(self, data):
        self.on_message(data)

    def vue_on_connected(self, connected):
        self.on_connected(connected)


@reacton.component
def FileSystemWatcher(server_path: str, on_connected: Callable[[bool], None], on_message: Callable[[Any], None], dev: bool):
    return FileSystemWatcherWidget.element(server_path=server_path, on_connected=on_connected, on_message=on_message, dev=dev)



import os
from typing import Any, Callable, Dict, Tuple, cast

import pandas as pd
import reacton
import reacton.ipyvuetify as v
import solara as sol
from solara.components.file_drop import FileDrop

_replacements = reacton.create_context(cast(Tuple[Dict[str, Any], Callable[[str, Any], None]], ({}, lambda key, value: None)))


def get_replacement(name: str) -> Any:
    if not reacton.core.get_render_context(required=False):
        return None
    replacements, _ = reacton.get_context(_replacements)
    return replacements.get(name)


def use_set_replacement():
    _, set_ = reacton.get_context(_replacements)
    return set_


class ReactiveDf:
    def __init__(self, var_name: str, initial_df_fn: Callable[[], pd.DataFrame], label: str):
        self.var_name = var_name
        self.initial_df_fn = initial_df_fn
        self.label = label

        if get_replacement(var_name) is None:
            self.df = initial_df_fn()
        else:
            self.df = get_replacement(var_name)

    def get(self):
        return self.df


@reacton.component
def DropDataframe(var_name: str, label: str):
    progress, set_progress = reacton.use_state(0)
    set_replacement = use_set_replacement()

    def on_file(file):
        fn_name = file["name"].lower().split(".")[-1]
        new_df = getattr(pd, f"read_{fn_name}")(file["file_obj"])
        set_replacement(var_name, new_df)

    server_software = os.environ.get("SERVER_SOFTWARE", "")
    is_in_solara = server_software.startswith("solara")

    with sol.Div() as main:
        FileDrop(
            label=label if is_in_solara else label + " (Note: does not work in preview mode)",
            on_total_progress=set_progress,
            on_file=on_file,
        )
        v.ProgressLinear(value=progress)

    return main



import os

import ipyvuetify as vy
import solara
import traitlets

images = [
    solara.util.load_file_as_data_url(os.path.join(os.path.dirname(__file__), name), "image/png")
    for name in [
        "hover_cell.png",
        "hover_icon.png",
        "menu.png",
        "load_data.png",
        "code_inserted.png",
        "menu_transform.png",
        "transform.png",
        "menu_visualizations.png",
        "visualizations.png",
        "menu_app.png",
        "app.png",
    ]
]


class TourWidget(vy.VuetifyTemplate):
    template_file = (__file__, "tour.vue")

    texts = traitlets.List(
        default_value=[
            "To start using Code Assist, hover over a code cell and the Code Assist button will appear.",
            "Hover over the Domino icon to open the Code Assist menu",
            "The menu items are in the order of a general data science workflow. First, you can load data.",
            "Each menu item opens a dialog that guides you through the process and generates code for you.",
            (
                "The generated code is inserted into the notebook and executed. The output of this, can be the input of the next step in the menu."
                " You can also edit or add to the code if you want."
            ),
            "Next we can transform the data in Transformations",
            "Lets remove missing values",
            "Now we can visualize the data",
            "Now we can visualize the data",
            "And finally we can create an app",
            "And finally we can create an app",
        ]
    ).tag(sync=True)
    images = traitlets.List(default_value=images).tag(sync=True)
    widths = traitlets.List(default_value=[722, 722, 722, 706, 804, 226.5, 790, 225, 950, 225.5, 853]).tag(sync=True)
    step = traitlets.Int(0).tag(sync=True)
    opened = traitlets.Bool(False).tag(sync=True)



import json
import threading
import time
import uuid
from urllib.parse import quote

import requests
from jupyter_server.base.handlers import JupyterHandler

from domino_code_assist import __version__
from domino_code_assist.domino_api import get_domino_api

EVENT_NAME = "LCA show button"


class MixpanelHandler(JupyterHandler):
    async def get(self, path=None):
        unique_id = "na"
        info = get_domino_api().get_user_info()
        if info:
            unique_id = info["id"] if info else None

        self.set_header("Content-Type", "application/json")
        self.write(
            json.dumps(
                {
                    "id": unique_id,
                    "LCA_version": __version__,
                    "Domino_version": get_domino_api().get_domino_version(),
                    "event_name": EVENT_NAME,
                }
            )
        )

    async def post(self, *args):
        data = json.loads(self.request.body)
        self._send_event(EVENT_NAME, data)

    def _send_event(self, event, props):
        unique_id = "na"
        info = get_domino_api().get_user_info()
        if info:
            unique_id = info["id"] if info else None

        event_item = {
            "event": event,
            "properties": {
                **props,
                "time": int(round(time.time() * 1000)),
                "distinct_id": unique_id,
                "$insert_id": str(uuid.uuid4()),
                "LCA_version": __version__,
                "Domino_version": get_domino_api().get_domino_version(),
                "$user_id": unique_id,
            },
        }

        def run():
            try:
                requests.post(
                    "https://api.mixpanel.com/track/",
                    headers={"content-type": "application/x-www-form-urlencoded"},
                    data=f"data={quote(json.dumps([event_item]))}",
                    timeout=3,
                )
            except Exception:
                pass

        threading.Thread(target=run).start()
